      program l2hcmp

c     Purpose: Compress the pixel dimensions of a pnm file by a factor.

c     Usage: See subroutine usage at the end of this file for
c            usage details.

c     Programming notes:

c     This program was written for the free GNU Fortran compiler g77
c     at www.gnu.org.  To make the l2hcomp.exe executable, use, say:
c     g77 -Wall -o ..\l2hcomp.exe l2hcomp.f

c     This uses a minimum dispersion algorithm, in which pixel fragments
c     are moved as little as possible, to preserve the graph integrity
c     as much as possible.

c     In the redistribution of pixel fragments, the starting point of
c     the current input pixel is kept inside the current output pixel,
c     so the output pixel can only stick out into the immediately
c     neighboring pixels.

c     avoid typos
      implicit none

c common:

c     debug option
      character*256 debug
      integer debugl
c     input file
      character*256 inf
      integer infl
c     output file
      character*256 outf
      integer outfl
c     compress option
      character*256 comp
      integer compl

c     common block
      common debugl,infl,outfl,compl,debug,inf,outf,comp

c local:

c     codes for the pnm formats recognized
      integer pbm,pgm,ppm
      parameter (pbm=4,pgm=5,ppm=6)
      integer ppbm,ppgm,pppm
      parameter (ppbm=1,ppgm=2,pppm=3)
c     actual format
      integer pid
c     desired output format
      integer pidw
      data pidw/0/

c     input picture width
      integer imax
c     index in the width direction
      integer i

c     input picture height
      integer jmax
c     index in the height direction
      integer j

c     input maximum color value
      integer cmax

c     compression factor
      real cmpfac

c     input pixel values as character indices
      integer icr,icg,icb
c     input pixel values as real numbers
      integer ir,ig,ib

c     declared output picture width
      integer iwdim
      parameter(iwdim=65536)
c     actual output picture width
      integer iwmax
c     output index in the width direction
      integer iw

c     output picture height
      integer jwmax
c     output index in the height direction
      integer jw

c     output maximum color value
      integer cmaxw

c     output pixel values
      real r(iwdim),g(iwdim),b(iwdim)
c     output pixel values of the next row of pixels
      real rn(iwdim),gn(iwdim),bn(iwdim)

c     picture width as real, equal to imax
      real width
c     picture height as real, equal to jmax
      real height

c     corresponding output pixel size
      real dxw,dyw

c     position of the left input pixel boundary
      real x2
c     position of the bottom input pixel boundary
      real y2
c     position of the left output pixel boundary
      real xw2
c     position of the bottom output pixel boundary
      real yw2

c     fraction of the input pixel sticking into neighboring output pixels
      real fracni,fracnj

c     scale factor from input to output pixel values
      real scale

c     number of command arguments
      integer iargc
      external iargc
      integer argmax
c     argument index
      integer arg
c     argument string
      character*256 argstr
c     argument length and index
      integer argl,argi

c     whether a file exist
      logical exists

c     number as string
      character*20 cnum

c     iteration index
      integer iter

c     pbm bit counter
      integer bit

c     input file character
      character*1 chr

c     error message
      character*66 errmsg

c executable:

c initialize the common block
      debugl=0
      infl=0
      outfl=0
      compl=0

c process the command line parameters:

c     read the command line parameters and process them

c     get number of arguments
      argmax=iargc()
      if(argmax.eq.0)call usage(' ')
      if(argmax.lt.4)call usage('too few parameters specified; need 4')
      if(argmax.gt.4)call usage('too many parameters specified; need 4')

      do 800 arg=1,argmax

c        get the argument
         argstr=' '
         call getarg(arg,argstr)
c        eliminate trailing whitespace
         do 50 argi=256,1,-1
            argl=argi
            if(argstr(argl:argl).gt.' ')goto 51
 50      continue
         call usage('parameter '//char(arg+48)//' is blank')
c        convert to lowercase
 51      if(arg.eq.2 .or.arg.eq.3)goto 61
         do 60 argi=1,argl
            if(argstr(argi:argi).ge.'A' .and. argstr(argi:argi).le.'Z')
     &       argstr(argi:argi)=char(ichar(argstr(argi:argi))+32)
 60      continue
 61      continue

c        process the parameter

         if(arg.eq.1)then

c           argument 1 is the debug option

c           set the debug option
            debugl=argl
            debug(1:debugl)=argstr(1:argl)
            call out(debug(1:debugl),'Debug option: '//debug(1:debugl))
c           check it
            if(debug(1:debugl).ne.'-d' .and. debug(1:debugl).ne.'-n')
     &       call usage('debug parameter can either be "-d" or "-n"')

         elseif(arg.eq.2)then

c           argument 2 is the input file

c           set the input file name
            infl=argl
            inf(1:infl)=argstr(1:argl)
            call out(debug(1:debugl),'Input file: '//inf(1:infl))
c           see whether it exists
            errmsg='error looking for input file'
            inquire(file=inf(1:infl),exist=exists,err=9900)
            if(.not. exists)call usage('input file not found')
c           open it for reading
            errmsg='error opening the input file'
            open(1,file=inf(1:infl),
     &       status='old',form='unformatted',err=9900)

         elseif(arg.eq.3)then

c           argument 3 is the output file

c           set the output file name
            outfl=argl
            outf(1:outfl)=argstr(1:argl)
            call out(debug(1:debugl),'Output file: '//outf(1:outfl))
c           see whether an old output file exists, and delete it if it does
            errmsg='error looking for the output file'
            inquire(file=outf(1:outfl),exist=exists,err=9900)
            if(exists)then
               errmsg='error opening the old output file'
               open(2,file=outf(1:outfl),status='old',err=9900)
               errmsg='error deleting the old output file'
               close(2,status='delete',err=9900)
            endif
c           open the output file
            errmsg='error opening the output file'
            open(2,file=outf(1:outfl),
     &       status='unknown',form='unformatted',err=9900)

         elseif(arg.eq.4)then

c           argument 4 is the compression factor

c           set the compression factor
            compl=argl
            comp(1:compl)=argstr(1:argl)
            call out(debug(1:debugl),
     &       'Compression parameter: '//comp(1:compl))
c           check it
            if(comp(1:1).ne.'-')call usage(
     &       'compression factor must be preceeded by "-"')
            if(compl.lt.2)call usage('missing compression factor')
            errmsg='invalid compression factor'
            read(comp(2:compl),*,err=9900)cmpfac
            if(cmpfac.le.1.)call usage(
     &       'compression factor must be greater than 1')

         endif

 800  continue

c read the input file header:

c     read the initial P
      call getchr(chr)
      if(chr.ne.'P')call fatal('input file is not a pnm file')
      call out(debug(1:debugl),'Have P in input file')

c     read the type ID
      call getchr(chr)
      pid=ichar(chr)-48
      call outi(debug(1:debugl),'Have pnm ID: ',pid)
      if(pid.eq.ppbm .or. pid.eq.ppgm .or. pid.eq.pppm)call fatal(
     & 'plain pnm encoding not supported; who uses them?')
      if(pid.ne.pbm .and. pid.ne.pgm .and. pid.ne.ppm)call fatal(
     & 'unrecognized input file format')
      if(pid.eq.pbm)call out(debug(1:debugl),'Input format: pbm')
      if(pid.eq.pgm)call out(debug(1:debugl),'Input format: pgm')
      if(pid.eq.ppm)call out(debug(1:debugl),'Input format: ppm')

c     read the figure width
      imax=0
      call gobble(chr)
      if(chr.lt.'0' .or. chr.gt.'9')call fatal(
     & 'missing width in the input file')
 1310 imax=imax*10+ichar(chr)-48
      call getchr(chr)
      if(chr.ge.'0' .and. chr.le.'9')goto 1310
      call outi(debug(1:debugl),'Width: ',imax)
      if(chr.gt.' ' .or. imax.eq.0)call fatal(
     & 'corrupt width in the input file')

c     read the height
      jmax=0
      call gobble(chr)
      if(chr.lt.'0' .or. chr.gt.'9')call fatal(
     & 'missing height in the input file')
 1410 jmax=jmax*10+ichar(chr)-48
      call getchr(chr)
      if(chr.ge.'0' .and. chr.le.'9')goto 1410
      call outi(debug(1:debugl),'Height: ',jmax)
      if(chr.gt.' ' .or. jmax.eq.0)call fatal(
     & 'corrupt height specification in the input file')

c     read the maximum color value cmax
      cmax=1
      if(pid.eq.pbm)goto 1590
      cmax=0
      call gobble(chr)
      if(chr.lt.'0' .or. chr.gt.'9')call fatal(
     & 'missing maximum color value in the input file')
 1510 cmax=cmax*10+ichar(chr)-48
      if(cmax.gt.255)call fatal('program maximum color value exceeded')
      call getchr(chr)
      if(chr.ge.'0' .and. chr.le.'9')goto 1510
      if(chr.gt.' ' .or. cmax.eq.0)call fatal(
     & 'corrupt end of header in input file')
 1590 call outi(debug(1:debugl),'Maximum color value: ',cmax)

c     use an output format with 256 color levels for most accurate results
      pidw=pid
      if(pidw.eq.pbm)pidw=pgm
      if(pidw.eq.pgm)call out(debug(1:debugl),'Output format: pgm')
      if(pidw.eq.ppm)call out(debug(1:debugl),'Output format: ppm')
      cmaxw=255
      call outi(debug(1:debugl),'Maximum color value out: ',cmaxw)

c     compute the output size, getting as close as possible on aspect ratio
      if(imax.le.jmax)then
         iwmax=imax/cmpfac+.5
         jwmax=float(jmax*iwmax)/imax+0.5
      else
         jwmax=jmax/cmpfac+.5
         iwmax=float(imax*jwmax)/jmax+0.5
      endif
      call outi(debug(1:debugl),'Output picture width: ',iwmax)
      call outi(debug(1:debugl),'Output picture height: ',jwmax)
      if(iwmax.lt.1)call fatal('compressed width too small')
      if(jwmax.lt.1)call fatal('compressed height too small')
      if(iwmax.gt.iwdim-1)call fatal('program memory exceeded')

c     write the output file header

      call putchr('P')
      call putchr(char(48+pidw))
      call putchr(char(10))
 1710 format(i20)
      write(cnum,1710)iwmax
      do 1720 iter=1,20
         if(cnum(iter:iter).ge.'0' .and. cnum(iter:iter).le.'9')
     &    call putchr(cnum(iter:iter))
 1720 continue
      call putchr(' ')
      write(cnum,1710)jwmax
      do 1730 iter=1,20
         if(cnum(iter:iter).ge.'0' .and. cnum(iter:iter).le.'9')
     &    call putchr(cnum(iter:iter))
 1730 continue
      call putchr(char(10))
      if(pidw.eq.pbm)goto 1750
      write(cnum,1710)cmaxw
      do 1740 iter=1,20
         if(cnum(iter:iter).ge.'0' .and. cnum(iter:iter).le.'9')
     &    call putchr(cnum(iter:iter))
 1740 continue
      call putchr(char(10))
 1750 continue

c     set some parameters

c     graph size as real numbers
      width=imax
      height=jmax

c     output pixel size
      dxw=width/iwmax
      dyw=height/jwmax

c     color scaling factor to account for different pixel size and cmax
      scale=cmaxw/(cmax*dxw*dyw)

c read and compress the rest of the input file:

c     initialize the output pixel row index
      jw=1

c     set the cell boundary
      yw2=dyw

c     initialize the output cells
      do 2100 iw=1,iwmax
         r(iw)=0.
         g(iw)=0.
         b(iw)=0.
         rn(iw)=0.
         gn(iw)=0.
         bn(iw)=0.
 2100 continue

c     initialize the red and blue input color values
      ir=0.
      ib=0.

      do 3000 j=1,jmax

c        set the cell boundary
         y2=j

c        fraction of the input cells in the next row of output cells
         fracnj=max(y2-yw2,0.)

c        initialize the next row of output cells
         iw=1
         xw2=dxw

c        the bit of the input file character we are in if we are reading pbm
         bit=8

         do 2900 i=1,imax

c           set the cell boundary
            x2=i

c           fraction of the input cell in the next output cell
            fracni=max(x2-xw2,0.)

c           get the next pixel from the file

            if(pid.eq.pbm)then

c              for PBM (B & W), read the blackness out of the correct bit
               bit=bit+1
               if(bit.gt.8)then
                  call getchr(chr)
                  bit=1
               endif
               icg=ichar(chr)/2**(8-bit)
               icg=icg-icg/2*2
               icg=cmax-icg
               ig=icg

            elseif(pid.eq.pgm)then

c              for PPM (grey), read the lightness
               call getchr(chr)
               icg=ichar(chr)
               if(icg.gt.cmax)
     &          call fatal('pixel exceeds maximum color value')
               ig=icg

            elseif(pid.eq.ppm)then

c              for PPM, read the RGB values
               call getchr(chr)
               icr=ichar(chr)
               call getchr(chr)
               icg=ichar(chr)
               call getchr(chr)
               icb=ichar(chr)
               if(icr.gt.cmax .or. icg.gt.cmax .or. icb.gt.cmax)
     &          call fatal('color exceeds maximum color value')
               ir=icr
               ig=icg
               ib=icb

            else

               call fatal('program error in read')

            endif

c           store the pixel value
            g(iw)=g(iw)+ig*scale*(1.-fracni)*(1.-fracnj)
            if(fracnj.gt.0.)gn(iw)=gn(iw)+ig*scale*(1.-fracni)*fracnj
            if(pidw.eq.ppm)then
               r(iw)=r(iw)+ir*scale*(1.-fracni)*(1.-fracnj)
               b(iw)=b(iw)+ib*scale*(1.-fracni)*(1.-fracnj)
               if(fracnj.gt.0.)then
                  rn(iw)=rn(iw)+ir*scale*(1.-fracni)*fracnj
                  bn(iw)=bn(iw)+ib*scale*(1.-fracni)*fracnj
               endif
            endif
            if(fracni.gt.0.)then
               g(iw+1)=g(iw+1)+ig*scale*fracni*(1.-fracnj)
               if(fracnj.gt.0.)gn(iw+1)=gn(iw+1)+ig*scale*fracni*fracnj
               if(pidw.eq.ppm)then
                  r(iw+1)=r(iw+1)+ir*scale*fracni*(1.-fracnj)
                  b(iw+1)=b(iw+1)+ib*scale*fracni*(1.-fracnj)
                  if(fracnj.gt.0.)then
                     rn(iw+1)=rn(iw+1)+ir*scale*fracni*fracnj
                     bn(iw+1)=bn(iw+1)+ib*scale*fracni*fracnj
                  endif
               endif
            endif

c           move to the next output pixel column as needed
 2890       if(x2.lt.xw2)goto 2900
            iw=iw+1
            xw2=iw*dxw
            goto 2890

 2900    continue

c        write out finished output rows
 2920    if(y2.lt.yw2 .and. j.lt.jmax)goto 3000

c        write the row jw
         if(pidw.eq.pgm)then
            do 2950 iw=1,iwmax
               icg=g(iw)+.5
               call putchr(char(icg))
 2950       continue
         else
            do 2960 iw=1,iwmax
               icr=r(iw)+.5
               call putchr(char(icr))
               icg=g(iw)+.5
               call putchr(char(icg))
               icb=b(iw)+.5
               call putchr(char(icb))
 2960       continue
         endif
         if(jw.eq.jwmax)goto 3010

c        proceed to the next row
         jw=jw+1
         yw2=jw*dyw
         do 2970 iw=1,iwmax
            r(iw)=rn(iw)
            g(iw)=gn(iw)
            b(iw)=bn(iw)
            rn(iw)=0.
            gn(iw)=0.
            bn(iw)=0.
 2970    continue

         goto 2920

 3000 continue
 3010 continue

c     done
 9000 call exit(100)

c     error exit
 9900 call fatal(errmsg)

      end



      subroutine getchr(chr)

c     Gets the next character from the input file.

c     avoid typos
      implicit none

c output:

c     character read
      character*1 chr

c local:

c     number of characters read
      integer ccnt
      data ccnt/0/

c     error status
      integer stat

c executable:

c     get the character
      call fgetc(1,chr,stat)

c     check for problems
      if(stat.eq.-1)then
         call outi('-d',
     &    '             Character count: ',ccnt)
         call fatal('premature end of input file')
      endif
      if(stat.ne.0)then
         call outi('-d',
     &    '             Character count: ',ccnt)
         call outi('-d',
     &    '             Input file I/O error code: ',stat)
         call fatal('error reading input file')
      endif

c     count the characters read so far
      ccnt=ccnt+1

c     all OK
      return
      end

      subroutine gobble(chr)

c     Gets the next character from the input file.  Gobbles white
c     space and comments.

c     avoid typos
      implicit none

c output:

c     character read
      character*1 chr

c local

c     number of characters in the current comment
      integer coml

c executable:

c     not in a comment yet
      coml=0

c     get the next character
 10   call getchr(chr)
      if(coml.gt.0 .and. chr.ne.char(10))goto 10
      coml=0
      if(chr.eq.'#')then
         coml=1
         goto 10
      endif
      if(chr.le.' ')goto 10

c     got it
      return
      end



      subroutine putchr(chr)

c     Write a character to the output file.

c     avoid typos
      implicit none

c input:

c     character to write
      character*1 chr

c local:

c     number of characters written
      integer ccnt
      data ccnt/0/

c     error status
      integer stat

c executable:

c     write the character
      call fputc(2,chr,stat)

c     check for problems
      if(stat.ne.0)then
         call outi('-d',
     &    '             Character count: ',ccnt)
         call outi('-d',
     &    '             Output file I/O error code: ',stat)
         call fatal('error writing to output file')
      endif

c     count the characters written so far
      ccnt=ccnt+1

c     all OK
      return
      end



      subroutine fatal(msg)

c     Prints the error message msg and aborts.

c     avoid typos
      implicit none

c input:

c     error message, 66 characters or less
      character*(*) msg

c common:

c     debug option
      character*256 debug
      integer debugl
c     input file
      character*256 inf
      integer infl
c     output file
      character*256 outf
      integer outfl
c     compression factor
      character*256 comp
      integer compl

c     common block
      common debugl,infl,outfl,compl,debug,inf,outf,comp

c executable:

c     print the message
      print10,char(7)
 10   format(a1,'*** l2hcomp: fatal error occurred')
      if(debugl.gt.0)
     & print20,'debug option: "',debug(1:min(debugl,50)),'"'
      if(infl.gt.0)
     & print20,'  input file: "',inf(max(1,infl-49):infl),'"'
      if(outfl.gt.0)
     & print20,' output file: "',outf(max(1,outfl-49):outfl),'"'
      if(compl.gt.0)
     & print20,' compression: "',comp(1:min(compl,50)),'"'
      print20,msg
 20   format('             ',3a)

c     wait a bit
      call sleep(5)

c     return error status 10
      call exit(10)
      end


      subroutine out(debug,msg)

c     Prints the message msg.

c     avoid typos
      implicit none

c input:

c     debug option
      character*(*) debug

c     message
      character*(*) msg

c executable:

c     print the message
      if(debug.eq.'-d')print10,msg
 10   format(a)

c     done
      return
      end

      subroutine outi(debug,msg,i)

c     Prints the message msg and the integer value i.

c     avoid typos
      implicit none

c input:

c     debug option
      character*(*) debug

c     message
      character*(*) msg

c     integer value
      integer i

c local

c     integer as string
      character*20 cnum
c     starting position
      integer start

c executable:

c     print the message and integer value
      write(cnum,10)i
 10   format(i20)
      do 20 start=1,19
         if(cnum(start:start).gt.' ')goto 21
 20   continue
 21   if(debug.eq.'-d')print50,msg,cnum(start:20)
 50   format(a,a)

c     done
      return
      end



      subroutine usage(msg)

c     Explains usage and aborts.

c     avoid typos
      implicit none

c input:

c     error message, 59 characters or less
      character*(*) msg

c common:

c     debug option
      character*256 debug
      integer debugl
c     input file
      character*256 inf
      integer infl
c     output file
      character*256 outf
      integer outfl
c     compression factor
      character*256 comp
      integer compl

c     common block
      common debugl,infl,outfl,compl,debug,inf,outf,comp

c executable:

c     print the message
      print10,'Usage:'
      print11,'l2hcomp -<debug> infile outfile -<factor>',
     & 'where:',
     & ' ',
     & '<debug> can be "d" for debugging output or "n" for not.',
     & ' ',
     & '<factor> is a compression factor greater than one.',
     & ' ',
     & 'On success, the return status is 100.  Everything else is',
     & 'failure.',
     & ' ',
     & '(If the above scrolled off your screen, use "l2hcomp '//
     & char(124)//' more")'
 10   format('*** l2hcomp: ',a)
 11   format('             ',a)

      if(debugl.le.0)goto 50
      print20
      print20,'Parameters specified:'
      print20, '   debug option: "',debug(1:min(debugl,50)),'"'
      if(infl.gt.0)
     & print20,'     input file: "',inf(max(1,infl-49):infl),'"'
      if(outfl.gt.0)
     & print20,'    output file: "',outf(max(1,outfl-49):outfl),'"'
      if(compl.gt.0)
     & print20,'    compression: "',comp(1:min(compl,50)),'"'
 20   format('             ',3a)
 50   if(msg.ne.' ')then
         print20
         print20,'Error: ',msg
      endif

c     wait a bit
      call sleep(5)

c     return error status 10
      call exit(10)
      end
