      program l2hcrp

c     Purpose: Crop and/or flip a latex2html or other pnm-format image
c              and examine whether the number of colors exceeds 256.

c     Usage: See subroutine usage at the end of this file for
c            usage details.

c     Programming notes:

c     This program was written for the free GNU Fortran compiler g77
c     at www.gnu.org.  To make the l2hcrop.exe executable, use, say:
c     g77 -Wall -o ..\l2hcrop.exe l2hcrop.f

c     To make a bigger version, increase bufdim and recompile.

c     (Unfortunately, g77 does not support the allocate statement.  Open
c     Watcom does, but it is several times slower in writing the output
c     file.  Also, g77 allocates local arrays of subroutines only on the
c     stack.  Conclusion: how big can images designed for a screen
c     really be?)

c     Pixel indices are i, varying in the horizontal direction from left
c     to right, and j, varying in the vertical direction from top to
c     bottom.  In the input file, pixels are stored with i varying most
c     rapidly.

c     Cutting is achieved by restricting the limits of i and j.

c     Flipping is achieved by writing the pixels to the output file
c     in the correctly revised order.

c     Internally, pixels are represented as RGB values, each restricted
c     to be in the range from 0 to "cmax", where cmax is found from the
c     input file and is not allowed to be more than 255, as is usual in
c     RGB color representation.  Thus each of the three RGB values can
c     be stored as a single 8 bit character.  For efficiency, however,
c     the RGB values are not stored directly.  What is stored is:
c
c     1) The RGB green value G.
c
c     2) The color "slope" S = G - R, where R is the RGB red value.  An
c        appropriate multiple of cmax+1 is added to keep S in the range
c        from 0 to cmax.
c
c     3) The color "nonlinearity" N = R + B - 2 G, where B is the RGB
c        blue value.  An appropriate multiple of cmax+1 is added to keep
c        N in the range from 0 to cmax.
c
c     Note that for each pixel, still three 8 bit characters are stored.
c     A big advantage of storing S and N instead of R and B is that in
c     grey regions, S and N are zero.
c
c     For black and white (pbm) and grey (pgm) formats, the S and N
c     characters are not needed.

c     All pixels are kept in memory.  Thus the input file has to be read
c     only once.

c     To be able to handle larger files, this program internally uses a
c     run length encoding to store the pixels.  The picture's G values
c     (and the S and N values the same way if ppm format) are stored as
c     strings of characters called "words".  Each word represents a
c     string of C consecutive characters and is stored in one of the
c     following four formats:
c
c     1) String of C nonrepeating characters.  The word's bit and
c        character pattern looks like:
c           [00uccccc] [character 1] [character 2] ... [character C]
c        where each [ ] denotes a single character, and ccccc are the
c        bits of the binary respesentation of C-1.  (Bit "u" is
c        currently undefined.  The idea is it may be used in a later
c        different encoding procedure, or to extend the maximum
c        possible picture width.)
c
c     2) Long string of C nonrepeating characters.  Another character
c        with bits for C-1 is added to the header:
c           [01uccccc] [cccccccc] [character 1] ... [character C]
c
c     3) String of a pixel character repeated C times:
c           [10uccccc] [repeated character]
c
c     4) Long string of a pixel character repeated C times:
c           [11uccccc] [cccccccc] [repeated character]
c
c     Note that in cases 3 and 4, only 2 or 3 characters are stored
c     to represent C characters, producing the reduction in storage.
c
c     Any string of 3 or more identical characters is represented by a
c     repeat count, 2 or less are left in "nonrepeating" words.  As a
c     result, the stored picture size cannot expand in size by any
c     significant amount even if rle is highly ineffective.  (Note that
c     pbm can still expand significantly from its file size, since
c     pixels are stored as bytes, not bits.)
c
c     Because up to 13 bits are available for C-1, the picture width
c     must be 2**13=8192 or less.  (Consecutive words of the same type
c     are not supported.)

c     avoid typos
      implicit none

c common:

c     debug option
      character*256 debug
      integer debugl
c     input file
      character*256 inf
      integer infl
c     output file
      character*256 outf
      integer outfl
c     crop options
      character*256 crop
      integer cropl
c     pad option
      character*256 padop
      integer padopl
c     flip option
      character*256 flip
      integer flipl
c     output format option
      character*256 fmtw
      integer fmtwl

c     common block
      common debugl,infl,outfl,cropl,padopl,flipl,fmtwl,
     & debug,inf,outf,crop,padop,flip,fmtw

c local:

c     codes for the pnm formats recognized
      integer pbm,pgm,ppm
      parameter (pbm=4,pgm=5,ppm=6)
      integer ppbm,ppgm,pppm
      parameter (ppbm=1,ppgm=2,pppm=3)
c     actual format
      integer pid
c     desired output format
      integer pidw
      data pidw/0/

c     declared picture width (above 2**13 = 8192 is not possible)
      integer wdim
      parameter(wdim=8192)
c     actual picture width
      integer width
c     index in the width direction
      integer i
c     cropped image limits
      integer imin,imax

c     declared picture height
      integer hdim
      parameter(hdim=8192)
c     actual picture height
      integer height
c     index in the height direction
      integer j
c     cropped image limits
      integer jmin,jmax

c     picture maximum color value (above 255 is not possible)
      integer cmax
c     picture maximum color value on output
      integer cmaxw
      data cmaxw/0/
c     RGB color values read in, on a scale from 0 to cmax
      integer icr,icg,icb
c     green value as character
      character*1 cg
c     color "slope" icg - icr (reduced to the range 0 to cmax)
      integer ics
c     "slope" value as character
      character*1 cs
c     color "nonlinearity"  icr - 2 icg + icb (reduced to the range 0 to cmax)
      integer icn
c     "nonlinearity" value as character
      character*1 cn
c     background color
      character*1 backcg,backcs,backcn
c     secondary background color
      character*1 bckcg2,bckcs2,bckcn2
c     character
      character*1 chr
c     current bit in the character in pbm format
      integer bit

c     pixel storage buffer size
      integer bufdim
      parameter (bufdim=750000)
c     pixel storage buffer
      character*1 buf(bufdim)
c     buffer size available for storing pixels
      integer buflim
c     buffer size actually used for storing pixels
      integer bufmax
c     buffer index
      integer bufi
c     starting buffer index for the encoded lines of G characters
      integer bufgs(hdim)
c     starting buffer index for the encoded lines of S characters
      integer bufss(hdim)
c     starting buffer index for the encoded lines of N characters
      integer bufns(hdim)

c     crop options index
      integer cropi

c     amounts cut from top, bottom, left, and right
      integer cutt,cutb,cutl,cutr
      data cutt/0/
c     amounts to pad top, botton, left, and right
      integer padt,padb,padl,padr
      data padt/0/,padb/0/,padl/0/,padr/0/

c     maximum pixels to remove in a shave
      integer shavmx
c     maximum bottom and side bar thicknesses
      integer barthb,barths

c     estimated length of a too short latex2html alignment bar
      integer barlt
c     estimated width of the bar
      integer barwd

c     empirical fudge factor; the number of pixels to add to the sidebar top
      integer fudge

c     declared number of color histogram color IDs
      integer hisdim
      parameter (hisdim=256)
c     maximum possible number of histogram color IDs
      integer hislim
c     actual number of histogram color IDs
      integer hismax
c     histogram color IDs
      integer hisid(hisdim)

c     graph corner index, clockwise with top left being 1
      integer corner
c     graph corner color IDs (with overflow)
      integer corid(7)
c     number of corners with the same color as the considered one
      integer coreq
c     the maximum value of coreq among all four corners
      integer coreqm
c     corresponding color ID
      integer coridm
c     count of corners with the maximum coreq value
      integer coreqc
c     whether a background color was found from the corner colors
      logical havbck
      data havbck/.false./

c     browser to optimize for
      character*1 bro
c     possible values
      character*7 bros
      parameter (bros='egmcaxf')
c     amount Mosaic raises an align=middle image above the baseline
      integer shiftm
c     amounts to pad top and bottom for correct alignment in Mosaic
      integer padtm,padbm
c     amount Explorer raises an align=middle image above the baseline
      integer shifte
c     amounts to pad top and bottom for correct alignment in Explorer
      integer padte,padbe
c     amounts in the previous iteration
      integer padte0,padbe0
c     height of the new image
      integer newh

c     estimated font size in pixels
      integer fntsiz
      data fntsiz/17/

c     size of the Mosaic shift data table
      integer mosdim
      parameter (mosdim=30)
c     shift for each font size in pixels
      integer mosshf(mosdim)
      data mosshf/
     & 0,0,1,1,1,1,2,2,2,2,
     & 3,3,3,4,4,4,4,5,5,5,
     & 5,6,6,6,6,7,7,7,7,8/
c     shift beyond mosdim = (font size in px) * mosslp + mosoff
      real mosslp,mosoff
      data mosslp/0.25/,mosoff/0.375/

c     size of the Internet Explorer shift data tables
      integer expdim
      parameter (expdim=36)
c     maximum shift for each font size in pixels
      integer expshf(expdim)
      data expshf/
     &  0, 1, 1, 1, 2, 2, 2, 3, 3, 3,
     &  4, 4, 5, 5, 5, 6, 5, 6, 7, 7,
     &  7, 7, 8, 7, 9, 9,10,10,10,10,
     & 10,11,12,12,12,12/
c     maximum shift beyond expdim = {font size in px) * expshs + expsho
      real expshs,expsho
      data expshs/0.3333333/,expsho/0./
c     actual maximum shift
      integer expsh
c     last font size for maximum shift *without rounding up*
      integer expsh1(expdim)
      data expsh1/
     &  1, 1, 2, 4, 2, 2, 4, 4, 6, 8,
     &  6, 8, 6, 6, 8, 6,10,10, 8,10,
     & 10,12,10,14,10,14,12,12,14,16,
     & 16,14,14,16,16,18/
c     beyond expdim = {font size in px) * exps1s + exps1o
      real exps1s,exps1o
      data exps1s/0.5/,exps1o/-0.5/
c     actual last font size for maximum shift *without rounding up*
      integer exps1
c     first font size for no shift *without rounding down*
      integer expsh2(expdim)
      data expsh2/
     &  3, 3, 6, 8,10,10,12,16,18,20,
     & 22,24,26,26,28,30,30,34,36,38,
     & 38,40,42,42,46,50,52,52,54,56,
     & 56,58,62,64,64,66/
c     beyond expdim = {font size in px) * exps2s + exps2o
      real exps2s,exps2o
      data exps2s/2./,exps2o/-8./
c     actual first font size for no shift *without rounding down*
      integer exps2

c     regression coefficients of the shift of the average style (a) browser
      real staoff,staslp
      data staoff/-0.3467284/,staslp/0.2779967/

c     size of the explorer style shift data table
      integer stxdim
      parameter (stxdim=41)
c     shift for each font size in pixels
      integer stxshf(stxdim)
      data stxshf/
     & 0,0,1,1,1,1,1,2,3,3,
     & 3,4,4,4,4,5,5,5,5,5,
     & 6,7,6,7,8,8,8,9,9,9,
     & 10,9,10,11,11,11,11,12,12,12,
     & 13/
c     regression coefficients of the shift of the Explorer style (x) browser
      real stxoff,stxslp
      data stxoff/-0.2799670/,stxslp/0.3191899/

c     size of the firefox style (f) browser shift data table
      integer stfdim
      parameter (stfdim=41)
c     shift for each font size in pixels
      integer stfshf(stfdim)
      data stfshf/
     & 0,0,0,0,1,1,1,2,2,2,
     & 2,2,3,3,3,3,4,4,4,4,
     & 5,5,5,5,6,6,6,6,6,7,
     & 7,7,7,8,8,8,8,9,9,9,
     & 9/
c     regression coefficients of the shift of the Firefox style (f) browser
      real stfoff,stfslp
      data stfoff/-0.4134897/,stfslp/0.2368035/

c     how much we increment i and j each time while writing the output file
      integer iinc,jinc
      data iinc/1/,jinc/1/
c     whether we still write the i-direction varying most rapidly after flip
      logical ifast
      data ifast/.true./

c     decoded output buffer height, width, and size
      integer bufh,bufw,bufsiz
c     i-range stored in the decoded buffer
      integer ibuflo,ibufhi
c     indices i and j for the flipped graph to be written to the output file
      integer iw,jw
c     corresponding maximum values
      integer iwmax,jwmax
c     mapping from iw and jw to i: i=io+iiw*iw+ijw*jw
      integer io,iiw,ijw
c     mapping from iw and jw to j: j=jo+jiw*iw+jjw*jw
      integer jo,jiw,jjw
c     buffer shift to account for overflow pixels
      integer bufish
c     mapping from iw and jw to bufi: bufi=bufio+bufiiw*iw+bufijw*jw
      integer bufio,bufiiw,bufijw
c     current buffer offset for the decoded pixels.
      integer bufioc

c     how many previous and following characters are equal to the current one
      integer prev,next

c     number of command arguments
      integer iargc
      external iargc
      integer argmax
c     argument index
      integer arg
c     argument string
      character*256 argstr
c     argument length and index
      integer argl,argi

c     position of the file extension point
      integer fextp

c     whether a file exist
      logical exists

c     iteration counter
      integer iter

c     number as string
      character*20 cnum

c     temporaries
      real temp
      integer itemp

c     error message
      character*66 errmsg

c executable:

c initialize the common block
      debugl=0
      infl=0
      outfl=0
      cropl=0
      padopl=0
      flipl=0
      fmtwl=0

c process the command line parameters:

c     read the command line parameters and process them

c     get number of arguments
      argmax=iargc()
      if(argmax.eq.0)call usage(' ')
      if(argmax.lt.7)call usage('too few parameters specified; need 7')
      if(argmax.gt.7)call usage('too many parameters specified; need 7')

      do 800 arg=1,argmax

c        get the argument
         argstr=' '
         call getarg(arg,argstr)
c        eliminate trailing whitespace
         do 50 argi=256,1,-1
            argl=argi
            if(argstr(argl:argl).gt.' ')goto 51
 50      continue
         call usage('parameter '//char(arg+48)//' is blank')
c        convert to lowercase
 51      if(arg.eq.2 .or.arg.eq.3)goto 61
         do 60 argi=1,argl
            if(argstr(argi:argi).ge.'A' .and. argstr(argi:argi).le.'Z')
     &       argstr(argi:argi)=char(ichar(argstr(argi:argi))+32)
 60      continue
 61      continue

c        process the parameter

         if(arg.eq.1)then

c           argument 1 is the debug option

c           set the debug option
            debugl=argl
            debug(1:debugl)=argstr(1:argl)

c           check it
            if(debug(1:debugl).ne.'-d' .and. debug(1:debugl).ne.'-n')
     &       call usage('debug parameter can either be "-d" or "-n"')

         elseif(arg.eq.2)then

c           argument 2 is the input file

c           set the input file name
            infl=argl
            inf(1:infl)=argstr(1:argl)
c           see whether it exists
            errmsg='error looking for input file'
            inquire(file=inf(1:infl),exist=exists,err=9900)
            if(.not. exists)call usage('input file not found')
c           open it for reading
            errmsg='error opening the input file'
            open(1,file=inf(1:infl),
     &       status='old',form='unformatted',err=9900)

         elseif(arg.eq.3)then

c           argument 3 is the output file, also used for the log file name

c           set the output file name
            outfl=argl
            outf(1:outfl)=argstr(1:argl)
c           see whether am old output file exists, and delete it if it does
            errmsg='error looking for the output file'
            inquire(file=outf(1:outfl),exist=exists,err=9900)
            if(exists)then
               errmsg='error opening the old output file'
               open(2,file=outf(1:outfl),status='old',err=9900)
               errmsg='error deleting the old output file'
               close(2,status='delete',err=9900)
            endif
c           open the output file
            errmsg='error opening the output file'
            open(2,file=outf(1:outfl),
     &       status='unknown',form='unformatted',err=9900)

c           put a tentative file extension point at the end of the file name
            fextp=outfl+1
            if(fextp.gt.256)call usage('output file name too long')
            outf(fextp:fextp)='.'
c           see whether there is already a file extension point in the name
            do 310 argi=outfl,1,-1
               if(outf(argi:argi).eq.'.')then
                  fextp=argi
                  goto 311
               endif
               if(outf(argi:argi).eq.'/')goto 311
               if(outf(argi:argi).eq.':')goto 311
               if(outf(argi:argi).eq.char(92))goto 311
 310        continue
 311        continue
c           see whether an old log file exists, and delete it if it does
            errmsg='error looking for the log file'
            inquire(file=outf(1:fextp)//'log',exist=exists,err=9900)
            if(exists)then
               errmsg='error opening the old log file'
               open(3,file=outf(1:fextp)//'log',status='old',err=9900)
               errmsg='error deleting the old log file'
               close(3,status='delete',err=9900)
            endif
c           open the log file
            errmsg='error opening the log file'
            open(3,file=outf(1:fextp)//'log',status='unknown',err=9900)

c           write the statistics so far
            call out(debug(1:debugl),'Debug option: '//debug(1:debugl))
            call out(debug(1:debugl),'Input file: '//inf(1:infl))
            call out(debug(1:debugl),'Output file: '//outf(1:outfl))
            call out(debug(1:debugl),'Log file: '//outf(1:fextp)//'log')

         elseif(arg.eq.4)then

c           argument 4 are the crop letters

c           set the crop option
            cropl=argl
            if(cropl.gt.255)call fatal('too many crop letters')
            crop(1:cropl+1)=argstr(1:argl)//' '
            call out(debug(1:debugl),'Crop option: '//crop(1:cropl))
c           check it
            if(crop(1:1).ne.'-')call usage(
     &       'crop letters must be preceeded by "-"')
            do 470 cropi=2,cropl
               if(index('tbrlas',crop(cropi:cropi)).le.0)call usage(
     &          'invalid crop character "'//crop(cropi:cropi)//'"')
 470        continue

         elseif(arg.eq.5)then

c           argument 5 is the pad option

c           set the pad option
            padopl=argl
            padop(1:padopl)=argstr(1:argl)
            call out(debug(1:debugl),'Pad option: '//padop(1:padopl))
c           check first character
            if(padop(1:1).ne.'-')call usage(
     &       'first character of the pad option must be -')

c           analyze the option

            if(
     &       padop(1:min(padopl,5)).eq.'-l2hb' .or.
     &       padop(1:min(padopl,5)).eq.'-l2hm')then

c              latex2html padding

c              get the desired browser type and font size, if given
               bro='a'
               if(padopl.gt.5)then
                  bro=padop(6:6)
                  if(index(bros,bro).gt.0)then
                     padop(6:6)=' '
                  else
                     bro='a'
                  endif
                  if(padop(6:padopl).ne.' ')then
                     read(padop(6:padopl),*,err=510)temp
                     if(temp.ge.5.)goto 520
 510                 call usage('browser or font size incorrect')
 520                 fntsiz=nint(temp)
                  endif
               endif

c              cut the specifications from the option
               padopl=5

c              no padding unless the crop options start with "abl"
               if(cropl.lt.4 .or. crop(1:4).ne.'-abl')then
                  call out(debug(1:debugl),'No latex2html padding '//
     &             'because of inconsistent cropping.')
                  padopl=1
               endif

            else if(padopl.gt.1)then

c              normal padding

c              check the color to use
               if(
     &          padop(2:2).ne.'w' .and.
     &          padop(2:2).ne.'b' .and.
     &          padop(2:2).ne.'c')call usage('the padding color '//
     &          'can only be "b", "w", or "c"')
c              initialize position in option string
               argi=2
c              read the next padding specification, if any
 530           if(argi.ge.padopl)goto 550
               itemp=0
c              read the next character in the specification
 540           argi=argi+1
               if(argi.gt.padopl)call usage('padding amounts '//
     &          'must be terminated by "t", "b", "l", or "r"')
               if(padop(argi:argi).ge.'0'
     &          .and. padop(argi:argi).le.'9')then
                  itemp=itemp*10+ichar(padop(argi:argi))-48
                  goto 540
               endif
               if(padop(argi:argi).eq.'t')then
                  if(padt.gt.0)call usage(
     &             'top padding specified twice')
                  padt=itemp
                  call outi(debug(1:debugl),'Top padding: ',padt)
               elseif(padop(argi:argi).eq.'b')then
                  if(padb.gt.0)call usage(
     &             'bottom padding specified twice')
                  padb=itemp
                  call outi(debug(1:debugl),'Bottom padding: ',padb)
               elseif(padop(argi:argi).eq.'l')then
                  if(padl.gt.0)call usage(
     &             'left padding specified twice')
                  padl=itemp
                  call outi(debug(1:debugl),'Left padding: ',padl)
               elseif(padop(argi:argi).eq.'r')then
                  if(padr.gt.0)call usage(
     &             'right padding specified twice')
                  padr=itemp
                  call outi(debug(1:debugl),'Right padding: ',padr)
               else
                  call usage('"'//padop(argi:argi)//
     &             '" is not a side to pad')
               endif
               goto 530
 550           continue

            endif

         elseif(arg.eq.6)then

c           argument 6 is the flip option

c           set the flip option
            flipl=argl
            flip(1:flipl)=argstr(1:argl)
            call out(debug(1:debugl),'Flip option: '//flip(1:flipl))

c           check the option and set output parameters
            if(flip(1:flipl).eq.'-xy'
     &       .or. flip(1:flipl).eq.'-transpose')then
c              for transpose, write out the figure with j varying fastest
               ifast=.false.
            else if(flip(1:flipl).eq.'-lr'
     &       .or. flip(1:flipl).eq.'-leftright')then
c              to flip left/right, i must go backwards
               iinc=-1
            elseif(flip(1:flipl).eq.'-tb'
     &       .or. flip(1:flipl).eq.'-topbottom')then
c              to flip top/bottom, j must go backwards
               jinc=-1
            elseif(flip(1:flipl).eq.'-cw' .or. flip(1:flipl).eq.'-r270'
     &       .or. flip(1:flipl).eq.'-rotate270')then
c              to rotate 270 degrees, j must go fastest and backwards
               ifast=.false.
               jinc=-1
            elseif(flip(1:flipl).eq.'-r180'
     &       .or. flip(1:flipl).eq.'-rotate180')then
c              to rotate 180 degrees, both i and j must go backwards
               iinc=-1
               jinc=-1
            elseif(flip(1:flipl).eq.'-ccw' .or. flip(1:flipl).eq.'-r90'
     &       .or. flip(1:flipl).eq.'-rotate90')then
c              to rotate 90 degrees, j must go fastest and i backwards
               ifast=.false.
               iinc=-1
            elseif(flip(1:flipl).eq.'-cxy'
     &       .or. flip(1:flipl).eq.'-ctranspose')then
c              for the other transpose, j must go fastest and both backwards
               ifast=.false.
               iinc=-1
               jinc=-1
            elseif(flip(1:flipl).ne.'-')then
               call usage('invalid flip option "'//
     &          flip(2:min(flipl,39))//'"')
            endif

         elseif(arg.eq.7)then

c           argument 7 is the output format

c           set the output format option
            fmtwl=argl
            fmtw(1:fmtwl)=argstr(1:argl)
            call out(debug(1:debugl),
     &       'Output format option: '//fmtw(1:fmtwl))
c           check first character
            if(fmtw(1:1).ne.'-')call usage(
     &       'first character of the output format option must be -')
            if(fmtwl.eq.1)goto 790
            argi=1
            if(fmtw(2:min(4,fmtwl)).eq.'pbm')then
               argi=argi+3
               pidw=pbm
            elseif(fmtw(2:min(4,fmtwl)).eq.'pgm')then
               argi=argi+3
               pidw=pgm
            elseif(fmtw(2:min(4,fmtwl)).eq.'ppm')then
               argi=argi+3
               pidw=ppm
            endif
            if(argi.ge.fmtwl)goto 790
            cmaxw=0
 750        argi=argi+1
            if(fmtw(argi:argi).lt.'0' .or. fmtw(argi:argi).gt.'9')
     &       call usage('invalid character "'//fmtw(argi:argi)//
     &       '" in output format')
            cmaxw=cmaxw*10+ichar(fmtw(argi:argi))-48
            if(argi.lt.fmtwl)goto 750
            if(cmaxw.eq.0)call usage(
     &       'the maximum color value cannot be 0')
            if(cmaxw.gt.255)call usage(
     &       'the maximum color value cannot exceed 255')
            if(pidw.eq.pbm .and. cmaxw.gt.1)call usage(
     &       'the maximum color value for pbm must be 1')
 790        if(pidw.eq.pbm)cmaxw=1

         endif

 800  continue

c     no latex2html padding for any flip option besides left/right
      if(padop(1:min(4,padopl)).eq.'-l2h')then
         if(
     &    flip(1:flipl).ne.'-' .and.
     &    flip(1:flipl).ne.'-leftright' .and.
     &    flip(1:flipl).ne.'-lr')then
            call out(debug(1:debugl),
     &       'No latex2html padding because of flip')
            padopl=1
         endif
      endif

c     write font size if the pad option is l2hm
      if(padop(1:min(5,padopl)).eq.'-l2hm')call outi(debug(1:debugl),
     & 'Image will be padded assuming a font size in px: ',fntsiz)

c     set fontsize-dependent parameters
      shavmx=nint(8./30.*fntsiz)
      barthb=nint(3.5/17.*fntsiz)
      barths=barthb

c read the input file header:

c     read the initial P
      call getchr(chr)
      if(chr.ne.'P')call fatal('input file is not a pnm file')
      call out(debug(1:debugl),'Have P in input file')

c     read the type ID
      call getchr(chr)
      pid=ichar(chr)-48
      call outi(debug(1:debugl),'Have pnm ID: ',pid)
      if(pid.eq.ppbm .or. pid.eq.ppgm .or. pid.eq.pppm)call fatal(
     & 'plain pnm encoding not supported; who uses them?')
      if(pid.ne.pbm .and. pid.ne.pgm .and. pid.ne.ppm)call fatal(
     & 'unrecognized input file format')
      if(pid.eq.pbm)call out(debug(1:debugl),'Input format: pbm')
      if(pid.eq.pgm)call out(debug(1:debugl),'Input format: pgm')
      if(pid.eq.ppm)call out(debug(1:debugl),'Input format: ppm')

c     read the figure width
      width=0
      call gobble(chr)
      if(chr.lt.'0' .or. chr.gt.'9')call fatal(
     & 'missing width in the input file')
 1310 width=width*10+ichar(chr)-48
      if(width.gt.wdim)call fatal('program width limit exceeded')
      call getchr(chr)
      if(chr.ge.'0' .and. chr.le.'9')goto 1310
      call outi(debug(1:debugl),'Width: ',width)
      if(chr.gt.' ' .or. width.eq.0)call fatal(
     & 'corrupt width in the input file')
      if(width.gt.2**13)call fatal('program error for wdim')

c     read the height
      height=0
      call gobble(chr)
      if(chr.lt.'0' .or. chr.gt.'9')call fatal(
     & 'missing height in the input file')
 1410 height=height*10+ichar(chr)-48
      if(height.gt.hdim)call fatal('program height limit exceeded')
      call getchr(chr)
      if(chr.ge.'0' .and. chr.le.'9')goto 1410
      call outi(debug(1:debugl),'Height: ',height)
      if(chr.gt.' ' .or. height.eq.0)call fatal(
     & 'corrupt file height specification in the input file')

c     read the maximum color value cmax
      cmax=1
      if(pid.eq.pbm)goto 1590
      cmax=0
      call gobble(chr)
      if(chr.lt.'0' .or. chr.gt.'9')call fatal(
     & 'missing maximum color value in the input file')
 1510 cmax=cmax*10+ichar(chr)-48
      if(cmax.gt.255)call fatal('program maximum color value exceeded')
      call getchr(chr)
      if(chr.ge.'0' .and. chr.le.'9')goto 1510
      if(chr.gt.' ' .or. cmax.eq.0)call fatal(
     & 'corrupt end of header in input file')
 1590 call outi(debug(1:debugl),'Maximum color value: ',cmax)

c     finalize the output format
      if(pid.eq.pbm .and. pidw.eq.0 .and. cmaxw.gt.1)pidw=pgm
      if(pidw.eq.0)pidw=pid
      if(cmaxw.eq.0)cmaxw=cmax
      if(pidw.eq.pbm)call out(debug(1:debugl),'Output format: pbm')
      if(pidw.eq.pgm)call out(debug(1:debugl),'Output format: pgm')
      if(pidw.eq.ppm)call out(debug(1:debugl),'Output format: ppm')
      call outi(debug(1:debugl),'Maximum color value out: ',cmaxw)

c read and encode the rest of the input file:

c     actual free space in the buffer excluding minimum working space
      buflim=bufdim-width
      if(pid.eq.ppm)buflim=bufdim-3*width
      if(buflim.le.0)then
         call outi('-d','Out of memory while reading row: ',0)
         call fatal('no more memory!')
      endif

c     last character used in the buffer
      bufmax=0

c     encode a line of zeros to allow pbm and pgm to be treated as ppm
      do 1610 i=1,width
         buf(bufdim-width+i)=char(0)
 1610 continue
      call encode(bufdim,bufmax,buflim,buf,bufdim-width,width,0)

c     store the rest of the file in rle encoding

      do 2000 j=1,height

c        the bit of the input file character we are in if we are reading pbm 
         bit=8

         do 1990 i=1,width

c           get the next pixel from the file

            if(pid.eq.pbm)then

c              for PBM (B & W), read the blackness out of the correct bit
               bit=bit+1
               if(bit.gt.8)then
                  call getchr(chr)
                  bit=1
               endif
               icg=ichar(chr)/2**(8-bit)
               icg=icg-icg/2*2
               icg=cmax-icg
               cg=char(icg)
               buf(bufdim-width+i)=cg

            elseif(pid.eq.pgm)then

c              for PPM (grey), read the lightness
               call getchr(chr)
               icg=ichar(chr)
               cg=chr
               if(icg.gt.cmax)
     &          call fatal('pixel exceeds maximum color value')
               buf(bufdim-width+i)=cg

            elseif(pid.eq.ppm)then

c              for PPM, read the RGB values
               call getchr(chr)
               icr=ichar(chr)
               call getchr(chr)
               icg=ichar(chr)
               cg=chr
               call getchr(chr)
               icb=ichar(chr)
               if(icr.gt.cmax .or. icg.gt.cmax .or. icb.gt.cmax)
     &          call fatal('color exceeds maximum color value')
               ics=mod((icg-icr)+cmax+1,cmax+1)
               cs=char(ics)
               icn=mod((icr+icb-2*icg)+2*cmax+2,cmax+1)
               cn=char(icn)
               buf(bufdim-3*width+i)=cn
               buf(bufdim-2*width+i)=cs
               buf(bufdim-width+i)=cg

            else

               call fatal('program error in read')

            endif

 1990    continue

c        encode the read line
         bufgs(j)=bufmax+1
         call encode(bufdim,bufmax,buflim,buf,bufdim-width,width,j)
         bufss(j)=1
         bufns(j)=1
         if(pid.eq.ppm)then
            bufss(j)=bufmax+1
            call encode(bufdim,bufmax,buflim,buf,bufdim-2*width,width,j)
            bufns(j)=bufmax+1
            call encode(bufdim,bufmax,buflim,buf,bufdim-3*width,width,j)
         endif

 2000 continue

      if(pid.eq.pbm)
     & call outi(debug(1:debugl),'Bits encoded: ',width*height)
      if(pid.eq.pgm)
     & call outi(debug(1:debugl),'Bytes encoded: ',width*height)
      if(pid.eq.ppm)
     & call outi(debug(1:debugl),'Bytes encoded: ',3*width*height)
      call outi(debug(1:debugl),'Encoded bytes: ',bufmax)

c cropping:

c     set the cropped limits to everything
      imin=1
      imax=width
      jmin=1
      jmax=height

c     initialize the crop character counter
      cropi=1

c     next crop
 3000 cropi=cropi+1
      if(cropi.gt.cropl)goto 3900
      if(crop(cropi:cropi).eq.'t')goto 3100
      if(crop(cropi:cropi).eq.'b')goto 3200
      if(crop(cropi:cropi).eq.'s')goto 3200
      if(crop(cropi:cropi).eq.'l')goto 3300
      if(crop(cropi:cropi).eq.'r')goto 3400
      if(crop(cropi:cropi).eq.'a')goto 3500
      call fatal('program error at crop letter')

c     do top cropping

c     find background color
 3100 call decode(bufdim,buf,bufgs(jmin),imin,imin,bufdim-1,prev,next)
      backcg=buf(bufdim)
      if(pid.ne.ppm)goto 3110
      call decode(bufdim,buf,bufss(jmin),imin,imin,bufdim-1,prev,next)
      backcs=buf(bufdim)
      call decode(bufdim,buf,bufns(jmin),imin,imin,bufdim-1,prev,next)
      backcn=buf(bufdim)
c     crop the all-background-color top rows
 3110 cutt=0
      do 3150 j=jmin,jmax
         call decode(bufdim,buf,bufgs(j),imin,imin,bufdim-1,prev,next)
         if(buf(bufdim).ne.backcg)goto 3190
         if(next.lt.imax-imin)goto 3190
         if(pid.ne.ppm)goto 3140
         call decode(bufdim,buf,bufss(j),imin,imin,bufdim-1,prev,next)
         if(buf(bufdim).ne.backcs)goto 3190
         if(next.lt.imax-imin)goto 3190
         call decode(bufdim,buf,bufns(j),imin,imin,bufdim-1,prev,next)
         if(buf(bufdim).ne.backcn)goto 3190
         if(next.lt.imax-imin)goto 3190
 3140    cutt=cutt+1
 3150 continue
      call fatal('cropped away entire graph from top')
 3190 jmin=jmin+cutt
      call outi(debug(1:debugl),'Crop '//crop(cropi:cropi)//
     & '; cut from top: ',cutt)
      if(crop(cropi:cropi).eq.'a')goto 3210
      goto 3000

c     do bottom cropping or shaving

c     find background color
 3200 call decode(bufdim,buf,bufgs(jmax),imin,imin,bufdim-1,prev,next)
      backcg=buf(bufdim)
      if(pid.ne.ppm)goto 3210
      call decode(bufdim,buf,bufss(jmax),imin,imin,bufdim-1,prev,next)
      backcs=buf(bufdim)
      call decode(bufdim,buf,bufns(jmax),imin,imin,bufdim-1,prev,next)
      backcn=buf(bufdim)
c     crop the all-background-color bottom rows
 3210 cutb=0
      do 3250 j=jmax,jmin,-1
         call decode(bufdim,buf,bufgs(j),imin,imin,bufdim-1,prev,next)
         if(buf(bufdim).ne.backcg)goto 3290
         if(next.lt.imax-imin)goto 3290
         if(pid.ne.ppm)goto 3240
         call decode(bufdim,buf,bufss(j),imin,imin,bufdim-1,prev,next)
         if(buf(bufdim).ne.backcs)goto 3290
         if(next.lt.imax-imin)goto 3290
         call decode(bufdim,buf,bufns(j),imin,imin,bufdim-1,prev,next)
         if(buf(bufdim).ne.backcn)goto 3290
         if(next.lt.imax-imin)goto 3290
 3240    if(crop(cropi:cropi).eq.'s' .and. cutb.ge.shavmx)goto 3290
         cutb=cutb+1
 3250 continue
      call fatal('cropped away entire graph from bottom')
 3290 jmax=jmax-cutb
      call outi(debug(1:debugl),'Crop '//crop(cropi:cropi)//
     & '; cut from bottom: ',cutb)
      if(crop(cropi:cropi).eq.'a')goto 3310
c     do we need a fix for the latex2html "alignment bar too short bug"?
      if(padop(1:min(padopl,4)).eq.'-l2h'
     & .and. cropi.eq.3 .and. cutb.eq.0)goto 3292
 3291 goto 3000
c     fix the  "bottom bar too short" or "side bar too long" bugs
 3292 call out('-d',
     & 'l2hcrop: attempting to fix an alignment-bar bug:')
c     find the actual bar length
      j=jmax
      call decode(bufdim,buf,bufgs(j),imin,imin,bufdim-1,prev,next)
      barlt=1+next
      if(pid.eq.ppm)then
         call decode(bufdim,buf,bufss(j),imin,imin,bufdim-1,prev,next)
         barlt=min(barlt,1+next)
         call decode(bufdim,buf,bufns(j),imin,imin,bufdim-1,prev,next)
         barlt=min(barlt,1+next)
      endif
c     find the true background color
      call decode(bufdim,buf,bufgs(j),imax,imax,bufdim-1,prev,next)
      bckcg2=buf(bufdim)
      if(pid.eq.ppm)then
         call decode(bufdim,buf,bufss(j),imax,imax,bufdim-1,prev,next)
         bckcs2=buf(bufdim)
         call decode(bufdim,buf,bufns(j),imax,imax,bufdim-1,prev,next)
         bckcn2=buf(bufdim)
      endif
c     see how wide the bar/true background color bottom strip is
      barwd=0
      do 3295 j=jmax,jmin,-1
         call decode(bufdim,buf,bufgs(j),imin,imin,bufdim-1,prev,next)
         if(buf(bufdim).ne.backcg .or. 1+next.lt.barlt)goto 3296
         call decode(bufdim,buf,bufgs(j),imax,imax,bufdim-1,prev,next)
         if(buf(bufdim).ne.bckcg2 .or. barlt+prev.lt.imax-imin)goto 3296
         if(pid.ne.ppm)goto 3294
         call decode(bufdim,buf,bufss(j),imin,imin,bufdim-1,prev,next)
         if(buf(bufdim).ne.backcs .or. 1+next.lt.barlt)goto 3296
         call decode(bufdim,buf,bufss(j),imax,imax,bufdim-1,prev,next)
         if(buf(bufdim).ne.bckcs2 .or. barlt+prev.lt.imax-imin)goto 3296
         call decode(bufdim,buf,bufns(j),imin,imin,bufdim-1,prev,next)
         if(buf(bufdim).ne.backcn .or. 1+next.lt.barlt)goto 3296
         call decode(bufdim,buf,bufns(j),imax,imax,bufdim-1,prev,next)
         if(buf(bufdim).ne.bckcn2 .or. barlt+prev.lt.imax-imin)goto 3296
 3294    barwd=barwd+1
 3295 continue
      call fatal('unable to fix the alignment bar bug: all is bar')
c     decide on what is the problem based on bar dimensions
 3296 if(barwd.le.0)then
         call outi('-d',
     &    'ignoring the b cut without a evident bug; barlt:',barlt)
         goto 3291
      else if(barwd.eq.1 .and. barlt.le.barths)then
c        assume side bar too long
         if(jmax.le.jmin)call fatal('cropped away entire graph')
         jmax=jmax-1
         call outi('-d',
     &    'shortened side alignment bar, bar thickness:',barlt)
         goto 3200
      else if(barwd.le.barthb .and. 2*barlt.ge.imax-imin)then
c        assume bottom bar too short
         call outi('-d',
     &    'extended bottom alignment bar, bar thickness:',barwd)
         cutb=barwd
         goto 3290
      else
c        just don't cut
         call outi('-d',
     &    'ignoring the b cut without a clear bug; barlt:',barlt)
         goto 3291
      endif

c     do left cropping

c     find background color
 3300 call decode(bufdim,buf,bufgs(jmax),imin,imin,bufdim-1,prev,next)
      backcg=buf(bufdim)
      if(pid.ne.ppm)goto 3310
      call decode(bufdim,buf,bufss(jmax),imin,imin,bufdim-1,prev,next)
      backcs=buf(bufdim)
      call decode(bufdim,buf,bufns(jmax),imin,imin,bufdim-1,prev,next)
      backcn=buf(bufdim)
c     initialize the bar length for the latex2html "bar too short" bug fix
      barlt=0
c     find the amount we can crop
 3310 cutl=imax-imin+1
      do 3350 j=jmax,jmin,-1
         call decode(
     &    bufdim,buf,bufgs(j),imin,imin,bufdim-1,prev,next)
         if(buf(bufdim).ne.backcg)next=-1
         cutl=min(cutl,next+1)
         if(pid.ne.ppm)goto 3340
         call decode(
     &    bufdim,buf,bufss(j),imin,imin,bufdim-1,prev,next)
         if(buf(bufdim).ne.backcs)next=-1
         cutl=min(cutl,next+1)
         call decode(
     &    bufdim,buf,bufns(j),imin,imin,bufdim-1,prev,next)
         if(buf(bufdim).ne.backcn)next=-1
         cutl=min(cutl,next+1)
 3340    if(cutl.le.0)goto 3390
         if(j.eq.jmax)barwd=cutl
         if(cutl.eq.barwd)barlt=barlt+1
 3350 continue
 3390 imin=imin+cutl
      if(imin.gt.imax)call fatal('cropped away entire graph from left')
      call outi(debug(1:debugl),'Crop '//crop(cropi:cropi)//
     & '; cut from left: ',cutl)
      if(crop(cropi:cropi).eq.'a')goto 3410
c     do we need a fix for the latex2html "alignment bar too short bug"?
      if(padop(1:min(padopl,4)).eq.'-l2h'
     & .and. cropi.eq.4 .and. cutl.eq.0)then
c        check whether it does look like an alignment bar
         if(barwd.ge.2 .and. barwd.le.4
     &    .and. 2*(barlt-1).ge.jmax-jmin)then
            cutl=barwd
            call out('-d',
     &       'l2hcrop: Fixed latex2html left alignment-bar bug:')
            call outi('-d','bar pixels missing: ',jmax-jmin+1-barlt)
            goto 3390
         endif
      endif
      goto 3000

c     do right cropping

c     find background color
 3400 call decode(bufdim,buf,bufgs(jmin),imax,imax,bufdim-1,prev,next)
      backcg=buf(bufdim)
      if(pid.ne.ppm)goto 3410
      call decode(bufdim,buf,bufss(jmin),imax,imax,bufdim-1,prev,next)
      backcs=buf(bufdim)
      call decode(bufdim,buf,bufns(jmin),imax,imax,bufdim-1,prev,next)
      backcn=buf(bufdim)
c     find the amount we can crop
 3410 cutr=imax-imin+1
      do 3450 j=jmin,jmax
         call decode(
     &    bufdim,buf,bufgs(j),imax,imax,bufdim-1,prev,next)
         if(buf(bufdim).ne.backcg)prev=-1
         cutr=min(cutr,prev+1)
         if(pid.ne.ppm)goto 3440
         call decode(
     &    bufdim,buf,bufss(j),imax,imax,bufdim-1,prev,next)
         if(buf(bufdim).ne.backcs)prev=-1
         cutr=min(cutr,prev+1)
         call decode(
     &    bufdim,buf,bufns(j),imax,imax,bufdim-1,prev,next)
         if(buf(bufdim).ne.backcn)prev=-1
         cutr=min(cutr,prev+1)
 3440    if(cutr.le.0)goto 3490
 3450 continue
 3490 imax=imax-cutr
      if(imax.lt.imin)call fatal('cropped away entire graph from right')
      call outi(debug(1:debugl),'Crop '//crop(cropi:cropi)//
     & '; cut from right: ',cutr)
      goto 3000

c     cut all sides with a uniform background or only find the background color

c     find a color ID for all 4 corners
 3500 do 3510 corner=1,4
         j=jmin
         if(corner.ge.3)j=jmax
         i=imin
         if(corner.eq.2 .or. corner.eq.3)i=imax
         call decode(bufdim,buf,bufgs(j),i,i,bufdim-1,prev,next)
         icg=ichar(buf(bufdim))
         corid(corner)=icg
         if(pid.ne.ppm)goto 3510
         call decode(bufdim,buf,bufss(j),i,i,bufdim-1,prev,next)
         ics=ichar(buf(bufdim))
         call decode(bufdim,buf,bufns(j),i,i,bufdim-1,prev,next)
         icn=ichar(buf(bufdim))
         icr=mod(icg-ics+cmax+1,cmax+1)
         icb=mod(icn-icr+2*icg+cmax+1,cmax+1)
         corid(corner)=icr*1000000+icg*1000+icb
 3510 continue
c     find the corner with the most neighboring corners with equal colors
      coreqm=0
      coridm=0
      coreqc=0
      corid(5)=corid(1)
      corid(6)=corid(2)
      corid(7)=corid(3)
      do 3520 corner=1,4
         coreq=1
         if(corid(corner).eq.corid(corner+1))coreq=coreq+1
         if(corid(corner).eq.corid(corner+2))coreq=coreq+1
         if(corid(corner).eq.corid(corner+3))coreq=coreq+1
         if(coreq.eq.coreqm)coreqc=coreqc+1
         if(coreq.le.coreqm)goto 3520
         coreqm=coreq
         coridm=corid(corner)
         coreqc=1
 3520 continue
      call outi(debug(1:debugl),
     & 'Maximum corners with equal color: ',coreqm)
      call outi(debug(1:debugl),
     & 'Number of such corner colors: ',coreqc/coreqm)
      call outi(debug(1:debugl),'Color ID of the first one: ',coridm)
c     set the corresponding background color
      if(pid.ne.ppm)then
         backcg=char(coridm)
      else
         icr=coridm/1000000
         icg=(coridm-icr*1000000)/1000
         icb=coridm-icr*1000000-icg*1000
         backcg=char(icg)
         backcs=char(mod(icg-icr+cmax+1,cmax+1))
         backcn=char(mod(icb-2*icg+icr+2*(cmax+1),cmax+1))
      endif
c     see whether the found background color is unique
      havbck=.false.
      if(coreqc/coreqm.eq.1)havbck=.true.
      if(.not.havbck)then
         call out(debug(1:debugl),
     &    '*** Cannot find a unique background color from the corners')
         goto 3000
      endif
c     done if we are simply looking at the corner colors
      if(crop(cropi:cropi).eq.'c')goto 3000
c     go cut all sides
      goto 3110

c     done cropping
 3900 continue

c padding

c     no padding
      if(padopl.le.1)goto 4900

c     padding with white
      if(padop(2:2).eq.'w')then
         backcg=char(cmax)
         backcs=char(0)
         backcn=char(0)
         goto 4900
      endif

c     padding with black
      if(padop(2:2).eq.'b')then
         backcg=char(0)
         backcs=char(0)
         backcn=char(0)
         goto 4900
      endif

c     first go do an "a" crop, or a "c" corner peek, if we did not do one yet
      cropi=cropl+1
      if(crop(cropi:cropi).ne.'a' .and. crop(cropi:cropi).ne.'c')then
c        set the crop letter
         crop(cropi:cropi)='a'
         if(padop(2:2).eq.'c')crop(cropi:cropi)='c'
c        undo any final shave crop for now
         if(crop(cropl:cropl).eq.'s')jmax=jmax+cutb
c        find the fudge factor in case this is an l2hm pad
         newh=jmax-jmin+1
         fudge=0
         if(fntsiz.eq.15)then
            if(cutt.le.1)then
               if(mod(newh,2).eq.0)fudge=-1
            else
               fudge=1
               if(mod(newh+fudge,2).eq.0)fudge=2
            endif
         else if(fntsiz.eq.17)then
            if(mod(newh,2).eq.0)fudge=1
         else if(fntsiz.eq.19)then
            if(cutt.le.1)then
               if(mod(newh,2).eq.0)fudge=-1
            else if(cutt.eq.2)then
               if(mod(newh,2).eq.0)fudge=1
            else
               fudge=1
               if(mod(newh+fudge,2).eq.0)fudge=2
            endif
         else
            if(cutt.le.1)then
               if(mod(newh,2).eq.0)fudge=-1
            else
               if(mod(newh,2).eq.0)fudge=1
            endif
         endif
c        this is another ugly empirical cludge
         if(bro.eq.'a' .and. fntsiz.eq.20)fudge=fudge+3
c        go do the crop
         goto 3500
      endif

c     reset things if the "a" or "c" crop failed
      if(.not.havbck)then
         if(crop(cropl:cropl).eq.'s')jmax=jmax-cutb
         if(padop(2:2).eq.'c')
     &    call fatal('no unique corner color found for padding')
         call out(debug(1:debugl),'Skipping latex2html padding '//
     &    'since no background color could be found.')
         goto 4900
      endif

c     case of simple padding
      if(padop(2:2).eq.'c')then
c        reset the shave, if any
         if(crop(cropl:cropl).eq.'s')jmax=jmax-cutb
         goto 4900
      endif

c     for latex2html padding, pad a pixel at each side
      padl=1
      padr=1

c     latex2html align=bottom padding
      if(padop(1:padopl).eq.'-l2hb')then
c        don't cut away the blank space below, say, raised superscripts
         if(cutb.gt.shavmx)jmax=jmax+cutb-shavmx
         goto 4900
      endif

c     perform latex2html align=middle padding

      if(padop(1:padopl).ne.'-l2hm')call fatal(
     & 'program error at l2hm padding')

c     to get Mosaic to align properly, we must have 2*shiftm more pixels
c     at the top of the figure

c     mosaic image shift above baseline
      if(fntsiz.le.mosdim)then
         shiftm=mosshf(fntsiz)
      else
         shiftm=nint(mosslp*fntsiz+mosoff)
      endif

c     represent the correct browser as a mosaic browser with shift 0
      if(bro.eq.'c')shiftm=0

c     average the generic browser between correct and IE
      if(bro.eq.'g')shiftm=0

c     set the average style browser
      if(bro.eq.'a')shiftm=nint(staoff+staslp*fntsiz)

c     set the explorer style browser
      if(bro.eq.'x')then
         shiftm=nint(stxoff+stxslp*fntsiz)
         if(fntsiz.le.stxdim)shiftm=stxshf(fntsiz)
      endif

c     set the firefox style browser
      if(bro.eq.'f')then
         shiftm=nint(stfoff+stfslp*fntsiz)
         if(fntsiz.le.stfdim)shiftm=stfshf(fntsiz)
      endif

c     log found shift
      call outi(debug(1:debugl),'Shiftm: ',shiftm)

c     try padding at the top only
      padbm=0
      padtm=2*shiftm+cutt-cutb+fudge
c     switch to bottom padding if needed
      if(padtm.lt.0)then
         padbm=-padtm
         padtm=0
      endif
      call outi(debug(1:debugl),'Fudge: ',fudge)
      call outi(debug(1:debugl),'Desired top padding: ',padtm)
      call outi(debug(1:debugl),'Desired bottom padding: ',padbm)

c     set optimal padding for mosaic
      padt=padtm
      padb=padbm

c     new height of the image
      newh=jmax-jmin+1+padt+padb

c     done if not optimizing for explorer or general
      if(index('eg',bro).le.0)goto 4890

c     IE cannot be aligned properly in general; its middle aligment is
c     affected by other images in the same line, something that is
c     unpredictable since it depends on the user's browser width
c     setting.  However, we can align hoping that there are no other
c     images in the same line.

c     data for IE shift above baseline
      if(fntsiz.le.expdim)then
         expsh=expshf(fntsiz)
         exps1=expsh1(fntsiz)
         exps2=expsh2(fntsiz)
      else
         expsh=nint(expshs*fntsiz+expsho)
         exps1=nint(exps1s*fntsiz+exps1o)
         exps2=nint(exps2s*fntsiz+exps2o)
      endif

c     guess shift based on mosaic image height
      if(newh.ge.exps2)then
         shifte=0
      else if(newh.le.exps1)then
         shifte=expsh
      else
         shifte=nint((newh-exps2)/(exps1-exps2+0.)*expsh)
      endif

c     since this shift depends on image height, we need to iterate
      do 4500 iter=1,6

c        try padding at the top only
         padbe=0
         padte=2*shifte+cutt-cutb+fudge
c        switch to bottom padding if needed
         if(padte.lt.0)then
            padbe=-padte
            padte=0
         endif

c        remember old paddings
         padte0=padte
         padbe0=padbe
c        average the paddings, rounding down to keep them minimal
         padt=(padtm+padte)/2
         padb=(padbm+padbe)/2
c        except if we are optimizing for exporer only
         if(bro.eq.'e')then
            padt=padte
            padb=padbe
         endif

c        new height of the image
         newh=jmax-jmin+1+padt+padb

c        compute a more accurate shift above baseline from the image height
         if(newh.ge.exps2)then
            shifte=0
         else if(newh.le.exps1)then
            shifte=expsh
         else
            shifte=nint((newh-exps2)/(exps1-exps2+0.)*expsh)
         endif

 4500 continue

c     if explorer keeps changing, be conservative
      if(bro.eq.'e')then
         padt=min(padte,padte0)
         padb=min(padbe,padbe0)
      endif

      call outi(debug(1:debugl),
     & 'Shift of image center above baseline: ',shifte)
      call outi(debug(1:debugl),
     & 'Desired top padding, revised: ',padte)
      call outi(debug(1:debugl),
     & 'Desired bottom padding, revised: ',padbe)

c     write statistics
 4890 call outi(debug(1:debugl),'New image height: ',newh)
      if(padt.ne.0)call outi(debug(1:debugl),'Top padding: ',padt)
      if(padb.ne.0)call outi(debug(1:debugl),'Bottom padding: ',padb)

c     end of padding
 4900 continue

c write the output with the appropriate flip option and padding

 8000 continue

c     find the dimensions of the figure to write out

      if(ifast)then
         iwmax=imax-imin+1+padl+padr
         jwmax=jmax-jmin+1+padt+padb
      else
         iwmax=jmax-jmin+1+padt+padb
         jwmax=imax-imin+1+padl+padr
      endif

c     write header

      call putchr('P')
      call putchr(char(48+pidw))
      call putchr(char(10))
 8010 format(i20)
      write(cnum,8010)iwmax
      do 8020 iter=1,20
         if(cnum(iter:iter).ge.'0' .and. cnum(iter:iter).le.'9')
     &    call putchr(cnum(iter:iter))
 8020 continue
      call putchr(' ')
      write(cnum,8010)jwmax
      do 8030 iter=1,20
         if(cnum(iter:iter).ge.'0' .and. cnum(iter:iter).le.'9')
     &    call putchr(cnum(iter:iter))
 8030 continue
      call putchr(char(10))
      if(pidw.eq.pbm)goto 8050
      write(cnum,8010)cmaxw
      do 8040 iter=1,20
         if(cnum(iter:iter).ge.'0' .and. cnum(iter:iter).le.'9')
     &    call putchr(cnum(iter:iter))
 8040 continue
      call putchr(char(10))
 8050 continue

c     dimension the decoded output buffer

      if(ifast)then

c        each i-direction output row comes from a single i-direction stored row

c        only one j=constant line needs to be decoded at a time
         bufh=1
c        decode all pixels in the j=constant line at a time
         bufw=iwmax
c        boost the number of pixels to a multiple of 8 for pbm
         if(pidw.eq.pbm)bufw=(bufw+7)/8*8
c        buffer size
         bufsiz=bufw
         if(bufsiz.gt.bufdim-bufmax)then
            call out('-d','Out of memory at the output stage!')
            call fatal('no more memory!')
         endif
         if((pidw.eq.ppm .or. pid.eq.ppm) .and.
     &    3*bufsiz.gt.bufdim-bufmax)then
            call out('-d','Out of memory at the output stage.')
            call fatal('no more memory!')
         endif
c        initialize overflow pixels
         if(pidw.eq.pbm)then
            do 8110 bufi=bufmax+1,bufmax+bufw
               buf(bufi)=char(cmaxw)
 8110       continue
         endif

      else

c        i-direction output rows contain pixels from every stored row

c        decode at least one pixel of every stored row
         bufh=iwmax
c        boost the number of pixels to a multiple of 8 for pbm
         if(pidw.eq.pbm)bufh=(bufh+7)/8*8
c        decode as much as possible from every stored row for efficiency
         bufw=(bufdim-bufmax)/bufh
         if(pidw.eq.ppm .or. pid.eq.ppm)bufw=(bufdim-bufmax)/(3*bufh)
         if(bufw.lt.1)then
            call out('-d','Out of memory at the output stage!')
            call fatal('no more memory!')
         endif
c        buffer size
         bufsiz=bufh*bufw
c        initialize overflow pixels
         if(pidw.eq.pbm)then
            do 8120 bufi=bufmax+iwmax*bufw+1,bufmax+bufsiz
               buf(bufi)=char(cmaxw)
 8120       continue
         endif

      endif

c     find the mapping from write coordinates to original coordinates

      if(ifast)then

c        i does not depend on jw
         ijw=0

c        set mapping from iw to i
         if(iinc.eq.1)then
            iiw=1
            io=imin-1-padl
         else
            iiw=-1
            io=imax+1+padr
         endif

c        j does not depend on iw
         jiw=0

c        set mapping from jw to j
         if(jinc.eq.1)then
            jjw=1
            jo=jmin-1-padt
         else
            jjw=-1
            jo=jmax+1+padb
         endif

      else

c        i does not depend on iw
         iiw=0

c        set mapping from jw to i
         if(iinc.eq.1)then
            ijw=1
            io=imin-1-padl
         else
            ijw=-1
            io=imax+1+padr
         endif

c        j does not depend on jw
         jjw=0

c        set mapping from iw to j
         if(jinc.eq.1)then
            jiw=1
            jo=jmin-1-padt
         else
            jiw=-1
            jo=jmax+1+padb
         endif

      endif

c     initialize actual number of colors in the histogram
      hismax=0

c     initialize maximum possible number of colors in the histogram
      hislim=min(cmaxw+1,cmax+1)**3
      if(pidw.eq.pbm .or. pid.eq.pbm)hislim=2
      if(pid.eq.pgm)hislim=min(hislim,cmax+1)
      if(pidw.eq.pgm)hislim=min(hislim,cmaxw+1)

c     initialize the decoded buffer as nothing stored
      ibuflo=1
      ibufhi=0

c     output each write row in turn

      do 8500 jw=1,jwmax

c        find the mapping from i and j to buffer index bufi

         if(ifast)then

c           buffer shift to get overflow pixels at end
            bufish=0
            if(pidw.eq.pbm .and.iinc.eq.-1)bufish=bufw-iwmax
c           decode the new row of pixels
            call decodp(bufdim,buf,hdim,bufgs,imin,imax,jmin,jmax,
     &       jo+jjw*jw,imin-padl,imax+padr,bufmax+bufish,backcg)
            if(pidw.eq.ppm .or. pid.eq.ppm)then
               call decodp(bufdim,buf,hdim,bufss,imin,imax,jmin,jmax,
     &          jo+jjw*jw,imin-padl,imax+padr,
     &          bufmax+bufish+bufsiz,backcs)
               call decodp(bufdim,buf,hdim,bufns,imin,imax,jmin,jmax,
     &          jo+jjw*jw,imin-padl,imax+padr,
     &          bufmax+bufish+2*bufsiz,backcn)
            endif

c           map iw and jw to bufi
            bufio=bufmax+bufish-imin+padl+1+io
            bufiiw=iiw
            bufijw=0

         else

c           decode more pixels if needed
            i=io+ijw*jw
            if(i.lt.ibuflo .or. i.gt.ibufhi)then
               ibuflo=i
               if(iinc.eq.-1)ibuflo=i-bufw+1
               ibufhi=ibuflo+bufw-1
               ibuflo=max(ibuflo,imin-padl)
               ibufhi=min(ibufhi,imax+padr)
               do 8200 iw=1,iwmax
                  call decodp(bufdim,buf,hdim,bufgs,
     &             imin,imax,jmin,jmax,jo+jiw*iw,ibuflo,ibufhi,
     &             bufmax+(iw-1)*bufw,backcg)
                  if(pidw.eq.ppm .or. pid.eq.ppm)then
                     call decodp(bufdim,buf,hdim,bufss,
     &                imin,imax,jmin,jmax,jo+jiw*iw,ibuflo,ibufhi,
     &                bufmax+(iw-1)*bufw+bufsiz,backcs)
                     call decodp(bufdim,buf,hdim,bufns,
     &                imin,imax,jmin,jmax,jo+jiw*iw,ibuflo,ibufhi,
     &                bufmax+(iw-1)*bufw+2*bufsiz,backcn)
                  endif
 8200          continue
            endif

c           map iw and jw to bufi
            bufio=bufmax-bufw+io-ibuflo+1
            bufiiw=bufw
            bufijw=ijw

         endif

c        find the buffer offset for the current write row
         bufioc=bufio+bufijw*jw

c        write the row

         if(pidw.eq.pbm)then

c           convert format, if needed
            if(pid.eq.ppm)then
c              convert back to RGB, then set the bit based on the color sum
               do 8230 iw=1,iwmax
                  icg=ichar(buf(bufioc+bufiiw*iw))
                  ics=ichar(buf(bufioc+bufiiw*iw+bufsiz))
                  icn=ichar(buf(bufioc+bufiiw*iw+2*bufsiz))
                  icr=mod(icg-ics+cmax+1,cmax+1)
                  icb=mod(icn-icr+2*icg+cmax+1,cmax+1)
                  if(icr+icg+icb.le.3*cmax/2)then
                     buf(bufioc+bufiiw*iw)=char(0)
                  else
                     buf(bufioc+bufiiw*iw)=char(1)
                  endif
 8230          continue
            elseif(pid.eq.pgm .and. cmax.ne.1)then
               do 8240 iw=1,iwmax
                  icg=ichar(buf(bufioc+bufiiw*iw))
                  if(icg.le.cmax/2)then
                     buf(bufioc+bufiiw*iw)=char(0)
                  else
                     buf(bufioc+bufiiw*iw)=char(1)
                  endif
 8240          continue
            endif

c           write the row compressed
            do 8250 iw=1,iwmax,8
               itemp=
     &          (1-ichar(buf(bufioc+bufiiw*(iw))))*128+
     &          (1-ichar(buf(bufioc+bufiiw*(iw+1))))*64+
     &          (1-ichar(buf(bufioc+bufiiw*(iw+2))))*32+
     &          (1-ichar(buf(bufioc+bufiiw*(iw+3))))*16+
     &          (1-ichar(buf(bufioc+bufiiw*(iw+4))))*8+
     &          (1-ichar(buf(bufioc+bufiiw*(iw+5))))*4+
     &          (1-ichar(buf(bufioc+bufiiw*(iw+6))))*2+
     &          (1-ichar(buf(bufioc+bufiiw*(iw+7))))
               call putchr(char(itemp))
               if(hismax.ge.hislim .or. hismax.gt.hisdim)goto 8250
               if(itemp.ne.0)call dohist(hisdim,hismax,hisid,0)
               if(iw+7.gt.iwmax)itemp=itemp+2**(iw+7-iwmax)-1
               if(itemp.ne.255)call dohist(hisdim,hismax,hisid,1)
 8250       continue

         elseif(pidw.eq.pgm)then

c           convert format, if needed
            if(pid.eq.ppm)then
c              convert back to RGB, then set grey based on the color sum
               do 8330 iw=1,iwmax
                  icg=ichar(buf(bufioc+bufiiw*iw))
                  ics=ichar(buf(bufioc+bufiiw*iw+bufsiz))
                  icn=ichar(buf(bufioc+bufiiw*iw+2*bufsiz))
                  icr=mod(icg-ics+cmax+1,cmax+1)
                  icb=mod(icn-icr+2*icg+cmax+1,cmax+1)
                  buf(bufioc+bufiiw*iw)=char(
     &             ((icr+icg+icb)*cmaxw+3*cmax/2)/(3*cmax))
 8330          continue
            elseif(cmaxw.ne.cmax)then
c              convert the color scale to cmaxw
               do 8340 iw=1,iwmax
                  icg=ichar(buf(bufioc+bufiiw*iw))
                  buf(bufioc+bufiiw*iw)=char((icg*cmaxw+cmax/2)/cmax)
 8340          continue
            endif

c           write the row plain
            do 8350 iw=1,iwmax
               call putchr(buf(bufioc+bufiiw*iw))
               if(hismax.ge.hislim .or.hismax.gt.hisdim)goto 8350
               call dohist(hisdim,hismax,hisid,
     &          ichar(buf(bufioc+bufiiw*iw)))
 8350       continue

         elseif(pidw.eq.ppm)then

c           decode s back into red, n back into blue
            do 8400 iw=1,iwmax
               icg=ichar(buf(bufioc+bufiiw*iw))
               ics=ichar(buf(bufioc+bufiiw*iw+bufsiz))
               icn=ichar(buf(bufioc+bufiiw*iw+2*bufsiz))
               icr=mod(icg-ics+cmax+1,cmax+1)
               icb=mod(icn-icr+2*icg+cmax+1,cmax+1)
               buf(bufioc+bufiiw*iw+bufsiz)=char(icr)
               buf(bufioc+bufiiw*iw+2*bufsiz)=char(icb)
 8400       continue

c           convert format, if needed
            if(cmaxw.ne.cmax)then
c              convert the color scale to cmaxw
               do 8440 iw=1,iwmax
                  icg=ichar(buf(bufioc+bufiiw*iw))
                  icr=ichar(buf(bufioc+bufiiw*iw+bufsiz))
                  icb=ichar(buf(bufioc+bufiiw*iw+2*bufsiz))
                  buf(bufioc+bufiiw*iw)=char((icg*cmaxw+cmax/2)/cmax)
                  buf(bufioc+bufiiw*iw+bufsiz)=
     &             char((icr*cmaxw+cmax/2)/cmax)
                  buf(bufioc+bufiiw*iw+2*bufsiz)=
     &             char((icb*cmaxw+cmax/2)/cmax)
 8440          continue
            endif

c           write the triplet row
            do 8450 iw=1,iwmax
               call putchr(buf(bufioc+bufiiw*iw+bufsiz))
               call putchr(buf(bufioc+bufiiw*iw))
               call putchr(buf(bufioc+bufiiw*iw+2*bufsiz))
               if(hismax.ge.hislim .or.hismax.gt.hisdim)goto 8450
               call dohist(hisdim,hismax,hisid,
     &          ichar(buf(bufioc+bufiiw*iw+bufsiz))*1000000+
     &          ichar(buf(bufioc+bufiiw*iw))*1000+
     &          ichar(buf(bufioc+bufiiw*iw+2*bufsiz)))
 8450       continue

         endif

 8500 continue

c     write the color statistics
      if(hismax.le.hisdim)then
         call outi(debug(1:debugl),'Number of colors: ',hismax)
         call outi(debug(1:debugl),'Lowest color ID: ',hisid(1))
         call outi(debug(1:debugl),'Highest color ID: ',hisid(hismax))
      else
         call outi(debug(1:debugl),'Number of colors exceeds: ',hisdim)
      endif

c     done
 9000 if(hismax.le.256 .and. min(iwmax,jwmax).ge.16)call exit(100)
      if(min(iwmax,jwmax).ge.16)call exit(101)
      if(hismax.le.256)call exit(102)
      call exit(103)

c     error exit
 9900 call fatal(errmsg)

      end



      subroutine encode(bufdim,bufmax,buflim,buf,bufoff,width,j)

c     Encodes the characters stored at bufoff+1, ... bufoff+width.
c     Stores the encoding starting at bufmax+1.  Increases bufmax to
c     indicate the buffer size now used.

c     avoid typos
      implicit none

c arguments:

c     declared buffer size
      integer bufdim

c     used buffer size
      integer bufmax

c     last available character in the buffer
      integer buflim

c     buffer
      character*1 buf(bufdim)

c     buffer offset for characters to encode
      integer bufoff

c     number of characters to encode
      integer width

c     row being encoded
      integer j

c local:

c     the character count represented by the current word
      integer cnt
c     the part of cnt that can be represented by the first word letter
      integer maxcnt
      parameter (maxcnt=32)
c     the overflow binary digits to be represented by the second letter
      integer over

c     the first character not yet stored
      integer buf0
c     the last character equal to the current one
      integer buf1
c     the current character
      integer bufi,bufi2

c executable:

c     initialize the first character not stored
      buf0=bufoff+1
c     initialize the last character equal to the current one
      buf1=bufoff+1
c     initialize the current character
      bufi=bufoff+1

c     read until we have 3 equal characters
 100  bufi=bufi+1
      if(bufi-bufoff.gt.width)then
         buf1=bufi
         goto 110
      endif
      if(buf(bufi).ne.buf(buf1))then
         buf1=bufi
         goto 100
      else
         if(bufi-buf1+1.lt.3)goto 100
      endif

c     store the string of unequal characters from buf0 to buf1-1, if any
 110  cnt=buf1-buf0
      if(cnt.gt.0)then
         if(cnt.le.maxcnt)then
            if(bufmax+cnt+1.gt.buflim)goto 900
            bufmax=bufmax+1
            buf(bufmax)=char(cnt-1)
         else
            if(bufmax+cnt+2.gt.buflim)goto 900
            over=(cnt-1)/maxcnt
            bufmax=bufmax+1
            buf(bufmax)=char(cnt-1-over*maxcnt+64)
            bufmax=bufmax+1
            buf(bufmax)=char(over)
         endif
         do 150 bufi2=buf0,buf1-1
            bufmax=bufmax+1
            buf(bufmax)=buf(bufi2)
 150     continue
      endif
      if(bufi-bufoff.gt.width)return
      buf0=buf1

c     read until we no longer have equal characters
 200  bufi=bufi+1
      if(bufi-bufoff.gt.width)goto 210
      if(buf(bufi).eq.buf(bufi-1))goto 200

c     store the repeated characters from buf1 to bufi-1
 210  cnt=bufi-buf1
      if(cnt.le.maxcnt)then
         if(bufmax+2.gt.buflim)goto 900
         bufmax=bufmax+1
         buf(bufmax)=char(cnt-1+128)
      else
         if(bufmax+3.gt.buflim)goto 900
         over=(cnt-1)/maxcnt
         bufmax=bufmax+1
         buf(bufmax)=char(cnt-1-over*maxcnt+64+128)
         bufmax=bufmax+1
         buf(bufmax)=char(over)
      endif
      bufmax=bufmax+1
      buf(bufmax)=buf(buf1)
      if(bufi-bufoff.gt.width)return
      buf0=bufi
      buf1=bufi
      goto 100

 900  call outi('-d','Out of memory at row: ',j)
      call fatal('no more memory!')
      end



      subroutine decode(bufdim,buf,bufs,i1,i2,bufoff,prev,next)

c     Decodes the characters between i1 and i2 from the encoded line
c     that starts at bufs, and stores the results at bufoff+1, ...
c     bufoff+i2-i1+1.  Also returns the number of characters previous to
c     the one at i1 that are equal to the character at i1 and the number
c     of characters beyond i2 that are equal to the one at i2.

c     avoid typos
      implicit none

c arguments:

c     declared buffer size
      integer bufdim

c     buffer
      character*1 buf(bufdim)

c     position where the encoded line starts
      integer bufs

c     starting and ending columns in the original line
      integer i1,i2

c     buffer offset to store the decoded characters
      integer bufoff

c     number of characters before i1 that are equal to it
      integer prev

c     number of characters after i2 that are equal to it
      integer next

c local:

c     current word's starting point
      integer bufw
c     next word's starting point
      integer bufwn

c     starting point of the characters described by the current word
      integer iw
c     starting point of the characters described by the next word
      integer iwn
c     running character index
      integer i

c     numerical value of the first letter of the current word
      integer first

c     the part of cnt-1 that can be represented by the first word letter
      integer maxcnt
      parameter (maxcnt=32)

c executable:

c     check parameters for sanity
      if(bufoff+i2-i1+1.gt.bufdim)
     & call fatal('program error: decode out of bounds')
      if(i2.lt.i1 .or. i1.le.0)
     & call fatal('program error: decode wrong bounds')

c     assume no previous and next identical characters
      prev=0
      next=0

c     initialize the next word position
      bufwn=bufs
      iwn=1

c     find the extend of the current word
 100  bufw=bufwn
      iw=iwn
      first=ichar(buf(bufw))
      if(first.ge.128+64)then
         iwn=iw+mod(first,maxcnt)+ichar(buf(bufw+1))*maxcnt+1
         bufwn=bufw+3
      elseif(first.ge.128)then
         iwn=iw+mod(first,maxcnt)+1
         bufwn=bufw+2
      elseif(first.lt.64)then
         iwn=iw+mod(first,maxcnt)+1
         bufwn=bufw+iwn-iw+1
      else
         iwn=iw+mod(first,maxcnt)+ichar(buf(bufw+1))*maxcnt+1
         bufwn=bufw+iwn-iw+2
      endif
      if(bufwn-1.gt.bufoff)call fatal(
     & 'program error: decode overwriting data')

c     next word if not yet reached the desired range
      if(i1.gt.iwn-1)goto 100

c     write the characters
      if(first.ge.128+64)then
         do 210 i=max(i1,iw),min(i2,iwn-1)
            buf(bufoff-i1+1+i)=buf(bufw+2)
 210     continue
         if(i1.ge.iw)prev=i1-iw
         if(i2.le.iwn-1)next=iwn-1-i2
      elseif(first.ge.128)then
         do 220 i=max(i1,iw),min(i2,iwn-1)
            buf(bufoff-i1+1+i)=buf(bufw+1)
 220     continue
         if(i1.ge.iw)prev=i1-iw
         if(i2.le.iwn-1)next=iwn-1-i2
      elseif(first.lt.64)then
         do 230 i=max(i1,iw),min(i2,iwn-1)
            buf(bufoff-i1+1+i)=buf(bufw-iw+1+i)
 230     continue
         if(i1.gt.iw)then
            if(buf(bufw-iw+1+i1).eq.buf(bufw-iw+i1))prev=1
         endif
         if(i2.lt.iwn-1)then
            if(buf(bufw-iw+1+i2).eq.buf(bufw-iw+2+i2))next=1
         endif
      else
         do 240 i=max(i1,iw),min(i2,iwn-1)
            buf(bufoff-i1+1+i)=buf(bufw-iw+2+i)
 240     continue
         if(i1.gt.iw)then
            if(buf(bufw-iw+2+i1).eq.buf(bufw-iw+1+i1))prev=1
         endif
         if(i2.lt.iwn-1)then
            if(buf(bufw-iw+2+i2).eq.buf(bufw-iw+3+i2))next=1
         endif
      endif

c     keep going until done
      if(iwn-1.lt.i2)goto 100
      return
      end



      subroutine decodp(
     & bufdim,buf,hdim,bufs,imin,imax,jmin,jmax,j,i1,i2,bufoff,back)

c     Version of decode that pads if it goes out of bounds.

c     avoid typos
      implicit none

c arguments:

c     declared buffer size
      integer bufdim

c     buffer
      character*1 buf(bufdim)

c     declared picture height
      integer hdim

c     positions where the encoded lines start
      integer bufs(hdim)

c     cropped image limits
      integer imin,imax,jmin,jmax

c     line to decode
      integer j

c     starting and ending columns in the original line
      integer i1,i2

c     buffer offset to store the decoded characters
      integer bufoff

c     background color to use for padding
      character*1 back

c local:

c     number of characters before i1 that are equal to it
      integer prev

c     number of characters after i2 that are equal to it
      integer next

c     running character index
      integer i

c executable:

c     check parameters for sanity
      if(bufoff+i2-i1+1.gt.bufdim)
     & call fatal('program error: decodp out of bounds')
      if(i2.lt.i1)
     & call fatal('program error: decodp wrong bounds')

c     handle the possibility of the entire line being out of bounds
      if(j.lt.jmin .or. j.gt.jmax .or. i2.lt.imin .or. i1.gt.imax)then
         do 100 i=i1,i2
            buf(bufoff-i1+1+i)=back
 100     continue
         return
      endif

c     put in the initial padding
      do 200 i=i1,imin-1
         buf(bufoff-i1+1+i)=back
 200  continue

c     decode the middle
      i=max(i1,imin)
      call decode(
     & bufdim,buf,bufs(j),i,min(imax,i2),bufoff+i-i1,prev,next)

c     put in the final padding
      do 300 i=imax+1,i2
         buf(bufoff-i1+1+i)=back
 300  continue

      return
      end



      subroutine getchr(chr)

c     Gets the next character from the input file.

c     avoid typos
      implicit none

c output:

c     character read
      character*1 chr

c local:

c     number of characters read
      integer ccnt
      data ccnt/0/

c     error status
      integer stat

c executable:

c     get the character
      call fgetc(1,chr,stat)

c     check for problems
      if(stat.eq.-1)then
         call outi('-d',
     &    '             Character count: ',ccnt)
         call fatal('premature end of input file')
      endif
      if(stat.ne.0)then
         call outi('-d',
     &    '             Character count: ',ccnt)
         call outi('-d',
     &    '             Input file I/O error code: ',stat)
         call fatal('error reading input file')
      endif

c     count the characters read so far
      ccnt=ccnt+1

c     all OK
      return
      end

      subroutine gobble(chr)

c     Gets the next character from the input file.  Gobbles white
c     space and comments.

c     avoid typos
      implicit none

c output:

c     character read
      character*1 chr

c local

c     number of characters in the current comment
      integer coml

c executable:

c     not in a comment yet
      coml=0

c     get the next character
 10   call getchr(chr)
      if(coml.gt.0 .and. chr.ne.char(10))goto 10
      coml=0
      if(chr.eq.'#')then
         coml=1
         goto 10
      endif
      if(chr.le.' ')goto 10

c     got it
      return
      end



      subroutine putchr(chr)

c     Write a character to the output file.

c     avoid typos
      implicit none

c input:

c     character to write
      character*1 chr

c local:

c     number of characters written
      integer ccnt
      data ccnt/0/

c     error status
      integer stat

c executable:

c     write the character
      call fputc(2,chr,stat)

c     check for problems
      if(stat.ne.0)then
         call outi('-d',
     &    '             Character count: ',ccnt)
         call outi('-d',
     &    '             Output file I/O error code: ',stat)
         call fatal('error writing to output file')
      endif

c     count the characters written so far
      ccnt=ccnt+1

c     all OK
      return
      end



      subroutine dohist(hisdim,hismax,hisid,colid)

c     Counts the number of histogram colors.

c     Note: parameter hismax must be initialized before the first call.

c     avoid typos
      implicit none

c parameters:

c     declared number of color histogram color IDs
      integer hisdim
c     current actual number of histogram color IDs
      integer hismax
c     histogram color IDs
      integer hisid(hisdim)
c     ID of the color to add to the histogram
      integer colid

c local:

c     position of the new color in the histogram
      integer hisi
c     lowest and highest possible value of hisi during search for it
      integer hisilo,hisihi

c executable:

c     skip histogram if already too many colors
      if(hismax.gt.hisdim)return

c     find the color ID place in the histogram by binary search
      hisilo=1
      hisihi=hismax+1
 100  if(hisilo.eq.hisihi)goto 200
      hisi=(hisilo+hisihi)/2
      if(colid.eq.hisid(hisi))return
      if(colid.lt.hisid(hisi))hisihi=hisi
      if(colid.gt.hisid(hisi))hisilo=hisi+1
      goto 100

c     put the new (since hisid(hisihi).ne.colid) color in the histogram
 200  hismax=hismax+1
      if(hismax.gt.hisdim)return
      do 300 hisi=hismax,hisilo+1,-1
         hisid(hisi)=hisid(hisi-1)
 300  continue
      hisid(hisilo)=colid

c     done with histogram
      return
      end



      subroutine fatal(msg)

c     Prints the error message msg and aborts.

c     avoid typos
      implicit none

c input:

c     error message, 66 characters or less
      character*(*) msg

c common:

c     debug option
      character*256 debug
      integer debugl
c     input file
      character*256 inf
      integer infl
c     output file
      character*256 outf
      integer outfl
c     crop options
      character*256 crop
      integer cropl
c     pad option
      character*256 padop
      integer padopl
c     flip option
      character*256 flip
      integer flipl
c     output format option
      character*256 fmtw
      integer fmtwl

c     common block
      common debugl,infl,outfl,cropl,padopl,flipl,fmtwl,
     & debug,inf,outf,crop,padop,flip,fmtw

c executable:

c     print the message
      print10,char(7)
 10   format(a1,'*** l2hcrop: fatal error occurred')
      if(debugl.gt.0)
     & print20,'debug option: "',debug(1:min(debugl,50)),'"'
      if(infl.gt.0)
     & print20,'  input file: "',inf(max(1,infl-49):infl),'"'
      if(outfl.gt.0)
     & print20,' output file: "',outf(max(1,outfl-49):outfl),'"'
      if(cropl.gt.0)
     & print20,' crop option: "',crop(1:min(cropl,50)),'"'
      if(padopl.gt.0)
     & print20,'  pad option: "',padop(1:min(padopl,50)),'"'
      if(flipl.gt.0)
     & print20,' flip option: "',flip(1:min(flipl,50)),'"'
      if(fmtwl.gt.0)
     & print20,'write format: "',fmtw(1:min(fmtwl,50)),'"'
      print20,msg
 20   format('             ',3a)

c     write it to the log file
      write(3,30,err=90)
 30   format('*** l2hcrop: fatal error occurred')
      write(3,20,err=90)msg
      write(3,20,err=90)'terminated on error'

c     exit with status 50 if we are out of memory
 90   if(msg.eq.'no more memory!')call exit(50)

c     wait a bit
      call sleep(5)

c     return error status 10
      call exit(10)
      end


      subroutine out(debug,msg)

c     Prints the message msg.

c     avoid typos
      implicit none

c input:

c     debug option
      character*(*) debug

c     message
      character*(*) msg

c executable:

c     print the message
      if(debug.eq.'-d')print10,msg
      write(3,10,err=90)msg
 10   format(a)

c     done
      return
 90   call fatal('unable to write to log file')
      end

      subroutine outi(debug,msg,i)

c     Prints the message msg and the integer value i.

c     avoid typos
      implicit none

c input:

c     debug option
      character*(*) debug

c     message
      character*(*) msg

c     integer value
      integer i

c local

c     integer as string
      character*20 cnum
c     starting position
      integer start

c executable:

c     print the message and integer value
      write(cnum,10)i
 10   format(i20)
      do 20 start=1,19
         if(cnum(start:start).gt.' ')goto 21
 20   continue
 21   if(debug.eq.'-d')print50,msg,cnum(start:20)
      write(3,50,err=90)msg,cnum(start:20)
 50   format(a,a)

c     done
      return
 90   call fatal('unable to write to log file')
      end



      subroutine usage(msg)

c     Explains usage and aborts.

c     avoid typos
      implicit none

c input:

c     error message, 59 characters or less
      character*(*) msg

c common:

c     debug option
      character*256 debug
      integer debugl
c     input file
      character*256 inf
      integer infl
c     output file
      character*256 outf
      integer outfl
c     crop options
      character*256 crop
      integer cropl
c     pad option
      character*256 padop
      integer padopl
c     flip option
      character*256 flip
      integer flipl
c     output format option
      character*256 fmtw
      integer fmtwl

c     common block
      common debugl,infl,outfl,cropl,padopl,flipl,fmtwl,
     & debug,inf,outf,crop,padop,flip,fmtw

c executable:

c     print the message
      print10,'Usage:'
      print11,'l2hcrop -<debug> infile outfile -<crops> -<pads> '//
     & '-<flip> -<fmt>',
     & 'where:',
     & ' ',
     & '<debug> can be "d" for debugging output or "n" for not.',
     & ' ',
     & '<crops> are one-character crops to be performed (or nothing):',
     & '   t: top,',
     & '   b: bottom,',
     & '   l: left,',
     & '   r: right,',
     & '   a: all four sides,',
     & '   s: shave one or two pixels from the bottom,',
     & 'Crops are only performed if all pixels involved are the same',
     & 'color.  An "a" crop crops the color shared by the most',
     & 'corners, if any.  If <pads> starts with "l2h", the b crop',
     & 'might be done crookedly to fix the "alignment bar too short"',
     & 'bug in latex2html, or the "side bar too long in ghostview".',
     & ' ',
     & '<pads> is the padding to be done, either nothing or one of:',
     & '   CN..NtN..NbN..NlN..Nr',
     & '   l2hb',
     & '   l2hm',
     & 'Here C is the color to pad the figure with: w for white, b',
     & 'for black, or c to use the color shared by the most corners.',
     & 'Further N..N stands for the amounts to pad at top, bottom,',
     & 'left, and right respectively.',
     & 'If <pads> is specified as l2hb or l2hm, cropping and padding',
     & 'for a latex2html "align=bottom", respectively "align=middle"',
     & 'image MAY be performed.  In particular, for latex2html padding',
     & 'to be performed, the first three letters of <crops> must be',
     & '"abl", and the only allowed <flip> is leftright or lr.  In',
     & 'addition, it must be possible to identify a unique background',
     & 'color as the color shared by the most corners, (ignoring a',
     & 'final "s" crop, if any.)  To optimize for specific browsers',
     & 'add one of mecgAEF behind l2hb or l2hm; an A is default.',
     & ' ',
     & '<flip> is the flip to perform (or nothing for no flip):',
     & '   rotate90, r90, ccw:  rotate the graph 90 degrees',
     & '   rotate180, r180:     rotate the graph 180 degrees',
     & '   rotate270, r270, cw: rotate the graph 90 degrees clockwise',
     & '   leftright, lr:       switch left and right',
     & '   topbottom, tb:       switch top and bottom',
     & '   transpose, xy:       switch bottom-left and top-right',
     & '   ctranspose, cxy:     switch bottom-right and top-left',
     & ' ',
     & '<fmt> is the output format:',
     & '       pbm: black and white',
     & '       pgm: grey scale',
     & '       ppm: RGB color (16M colors)',
     & '   nothing: to keep the format of the input file',
     & 'Any of the four above may directly be followed by the maximum',
     & 'color value (which is *not* the number of colors for ppm.)',
     & 'The specified maximum color value can range from 1 to 255.',
     & 'The file will be converted to the specified format by brute',
     & 'force, not cunning.  Specifying a maximum color value greater',
     & 'than 1 for pbm is an error; if a maximum color value greater',
     & 'than 1 is specified, but no output format, and the input file',
     & 'is pbm, it is converted to pgm on output.',
     & ' ',
     & 'Options are processed in command line order.',
     & ' ',
     & 'The program leaves a log file named according to outfile;',
     & 'e.g, cropping graph.pnm leaves a log file graph.log.',
     & ' ',
     & 'On success, the return status is:',
     & '   100: if the number of colors is less or equal to 256',
     & '   101: if the number of colors is greater than 256',
     & '   102: as 100 and minimum picture dimension less than 16',
     & '   103: as 101 and minimum picture dimension less than 16',
     & ' ',
     & 'If the program runs out of memory, it returns status 50.  In',
     & 'that case, run a bigger version.  Everything else is failure.',
     & ' ',
     & '(If the above scrolled off your screen, use "l2hcrop '//
     & char(124)//' more")'
 10   format('*** l2hcrop: ',a)
 11   format('             ',a)

      if(debugl.le.0)goto 50
      print20
      print20,'Parameters specified:'
      print20, '   debug option: "',debug(1:min(debugl,50)),'"'
      if(infl.gt.0)
     & print20,'     input file: "',inf(max(1,infl-49):infl),'"'
      if(outfl.gt.0)
     & print20,'    output file: "',outf(max(1,outfl-49):outfl),'"'
      if(cropl.gt.0)
     & print20,'    crop option: "',crop(1:min(cropl,50)),'"'
      if(padopl.gt.0)
     & print20,'     pad option: "',padop(1:min(padopl,50)),'"'
      if(flipl.gt.0)
     & print20,'    flip option: "',flip(1:min(flipl,50)),'"'
      if(fmtwl.gt.0)
     & print20,'  output format: "',fmtw(1:min(fmtwl,50)),'"'
 20   format('             ',3a)
 50   if(msg.ne.' ')then
         print20
         print20,'Error: ',msg
      endif

c     wait a bit
      call sleep(5)

c     return error status 10
      call exit(10)
      end
