      program subbin

c     Perform string substitutions in a file using file binary access.

c     Exit code is 1 on a fatal error, otherwise zero.

c     Keeps the file in memory, so is very fast, but can run out of
c     memory; g77 does not allow dynamic memory allocation.  Must be
c     compiled using g77 through the makefile.

c     Usage:  subbin substitutions-file old-file new-file [-v]
c             -v turns on verbosity, -V even more.

c     The text read from old-file is processed according to the contents
c     of substitutions-file and saved as new-file.  However, if there
c     are no changes to the file, no new file is produced.

c     Substitutions file must be of the form:
c        original_1
c        substitution_1
c        original_2
c        substitution_2
c        ...
c     Empty lines preceding the originals are ignored.
c
c     For each original/translation pair, the entire text is scanned and
c     every occurrence of the original is replaced by the translation.
c
c     Backslashes in either original or substitution are special; they
c     start commands, as described below.  To get a plain backslash,
c     double it as \\.  Lines are terminated by either linefeed (Unix)
c     or by carriage-control/linefeed (DOS).  A carriage-control
c     character anywhere else than before a linefeed is an error.
c
c     To translate an arbitrary binary string into another one, replace
c     all carriage returns with \M, all linefeeds with \J and all
c     backslashes with \\.  Then terminate each string with a newline.
c
c     Watch for invisible blanks at the end of lines or at end of the
c     file.  For example, a substitution file containing a single blank
c     space looks empty, but it will remove every blank space from
c     old-file.  Use the cursor keys to check for the presence of such
c     blanks.  Or select the entire file in you editor; usually this
c     highlights the text, allowing trailing blanks to be seen.

c Command reference

c     Commands describing special characters:
C
c        \\ is \
c        \  is ASCII nul (there must be a space behind the backslash)
c        \A through \Z are ASCII 1 to 26
c        \7 through \1 are ASCII 27 to 31
c        \2 is the delete character
c        \3 adds 128 to the ascii value of the next character
c     Note that newline for Windows/DOS format files is \M\J while
c     it is simply \J for Unix format files.

c     The next character selecting commands can only occur in the original:
c
c     command to test for end of file:
c        \] is end-of-file
c
c     commands governing case sensitivity:
c        \- turns off case sensitivity
c        \= turns on case sensitivity (default at the start of each original)
c
c     commands describing special allowed character specifications
c        \' is a single or double quote
c        \_ is a space or tab
c        \+ is a space, tab, carriage return, or line feed
c        \5 is any digit
c        \6 is a plus or minus sign
c        \a is any character, but not end of file
c        \l is any lower case letter
c        \u is any upper case letter (note lower case u)
c
c     commands to define a group of allowed character specifications
c        \& starts a group of match specifications (not using \&, \?, \~)
c        \* ends the group
c     aliases are \( and \)
c
c     command to invert allowed and nonallowed characters
c        \~ inverts the following matching
c
c     command to allow zero or more occurrences to count as a match:
c        \? is all following matching characters (zero or more)
c
c     Compound sequences must be ordered as one of:
c         [case change][\?][\~]allowed_character_specification
c         [case change][\?][\~]\&...\*
c     Case changes inside ... are fine and are not limited to the sequence.
c
c     For example, to match an integer number, use \?\+\?\6\5\?\5.  To
c     match everything until the next letter, use \-\?\~\u\=, or
c     \?\~\&\u\l\*, or even \?\~\&\-\l\=\*.

c     This uses a greedy, keep-on-trucking algorithm, where character
c     specifications for the original are pursued until they fail before
c     the next character specification is pursued.  It makes results
c     more predicable and the algorithm much faster.

c     To avoid infinite loops, substitutions that replace no characters
c     in the original file are only done once.  For example,
c        \?\+
c        -
c     puts a single hyphen before every nonwhite character in the file,
c     not infinitely many of them, including one before EOF, (as well as
c     the expected single hyphen replacing every contiguous nonzero
c     sequence of white space characters.)  Similarly,
c        \=
c        \J
c     puts a single linefeed before every character in the original file,
c     including before EOF, not infinitely many of them.

c     Special original strings:
c
c        \m  takes the original to be the current memory value.  If that
c            string is defined but nul, the original is taken as
c            \~\&\a\]\*, or in words, not anything.
c
c     The substitution should be omitted for the following special originals:
c
c        \r  Restart the substitutions file from the beginning
c        \x  Exit if there were no nontrivial substitutions since the last time
c
c        \d# unselects # levels of selection done by \s below, or unselects all
c            if # is omitted.

c     Special substitution strings:
c
c        \l converts the original to lower case.
c        \u converts the original to uppercase.
c        \x converts the first letter in the original to upper case.
c
c        \rcsss...ss replaces every character c in the original by character
c                    string sss...ss.  This string may be empty to delete c.
c        \esss...ssc encodes any character string sss..s in the original as
c                    the single character c, usually for further manipulation
c                    like with \r above. If a character like ASCII nul does
c                    not appear in the file, it makes a good choice for c.
c
c        \m  takes the substitution to be the current memory value.
c
c     In the following special substitution strings, # stands for an number
c     specifying a particular occurrence of the original string.  If the
c     number is omitted, the last occurence is used.  If there are less
c     occurences of the original string than the number #, a nul string at
c     the end of file is selected instead.
c
c        \c# does not modify the originals, but copies the #-th instance
c            to a memory location.
c        \f# simply finds the #-th occurrence.  (Not useful, included for
c            programming purposes.)
c        \s# does not modify the originals, but selects the #-th instance.
c            Further manipulations will be restricted to the selected string.
c            This is very powerful: the commands \l, \u, \x, \r, and \e above
c            can likely be done using \s looped with \x and \r.  However, that
c            would be much slower for multiple instances: each \s selection
c            from the full buffer requires the full buffer to be processed.
c            Be sure to deselect with \d before exiting.
c        \v# does not modify the originals, except converts the #-th instance
c            to lowercase.  (Kept for backward compatibility or maybe speed.)
c        \^# does not modify the originals, except converts the #-th instance
c            to uppercase.  (Kept for backward compatibility or maybe speed.)

c     Note: To address the entire file, original \a\?\a is best.  For example,
c     to change the file to upper case, use a substitution file:
c        \a\?\a
c        \^
c     Note that \?\a will be found twice, (first all characters will be found,
c     then zero characters at EOF will be found,) so \a\?\a is really needed.

c declarations

c     avoid typos
      implicit none

c     command line arguments
      integer argmx
      character*132 argt(4)
      integer argtl(4)
      external args

c     verbosity
      integer verbos

c executable

c     get the command line arguments
      call args(4,
     & 'Usage: subbin substitutions_file in_file out_file [-{v|V}]',
     & argmx,argt,argtl)
      verbos=0
      if(argmx.eq.4)then
         verbos=-1
         if(argtl(4).ne.2)goto 9990
         if(argt(4)(1:2).eq.'-v')verbos=1
         if(argt(4)(1:2).eq.'-V')verbos=2
         if(verbos.lt.0)goto 9990
      else if(argmx.ne.3)then
         goto 9990
      endif
      if(argtl(1).le.0 .or. argtl(2).le.0 .or. argtl(3).le.0)goto 9990

c     subbn0 does the actual work
      call subbn0(argt(1)(1:argtl(1)),argt(2)(1:argtl(2)),
     & argt(3)(1:argtl(3)),verbos)

c     all done
      call exit(0)

c     error exit
 9990 print9991,
     & 'Usage: subbin substitutions_file in_file out_file [-{v|V}]'
 9991 format(a)
      call fatal0
      end
